// This file is part of mzCalc.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License version 2, as published by the
// Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// mzCalc. If not, see <http://www.gnu.org/licenses/>.
//
// This work is Copyright (C) 2016 by Josse van der Plaat
//
// Individual portions may be copyright by individual contributors, and are
// included in this collective work with permission of the copyright owners.

#ifndef FLEET_H
#define FLEET_H

#include "battleformation.h"

#include "maybe.h"
#include "defines.h"
#include "verbosity.h"

#include <vector>

#include <cstdint>

struct Side {
    enum S {
        Attack = true,
        Defense = false
    } side;
    implicit Side(S s)
        : side(s)
    { }

    std::string toString() const {
        return side == Attack ? "Attack" : "Defend";
    }
    Side operator!() const {
        return Side(S(!side));
    }
};
bool operator==(const Side& lhs, const Side& rhs);
bool operator!=(const Side& lhs, const Side& rhs);
bool operator<( const Side& lhs, const Side& rhs);
bool operator>=(const Side& lhs, const Side& rhs);
bool operator<=(const Side& lhs, const Side& rhs);
bool operator>( const Side& lhs, const Side& rhs);


struct Fleet {
    explicit Fleet(int64_t index,
                   Side side)
        : index(index)
        , side(side)
    { }
    explicit Fleet(const Fleet& other)
        : index(other.index)
        , side(other.side)
        , xp(other.xp)
        , score(other.score)
        , value(other.value)
        , race(other.race)
        , militaryCenters(other.militaryCenters)
        , salvage(other.salvage)
        , roidsAllowedToCap(other.roidsAllowedToCap)
        , battleFormations(other.battleFormations)
    {
        for (BattleFormation& bf : battleFormations) {
            bf.fleet = this;
        }
    }
    Fleet& operator=(const Fleet& other) {
        index = other.index;
        side = other.side;
        xp = other.xp;
        score = other.score;
        value = other.value;
        race = other.race;
        militaryCenters = other.militaryCenters;
        salvage = other.salvage;
        roidsAllowedToCap = other.roidsAllowedToCap;
        for (BattleFormation& bf : battleFormations) {
            bf.fleet = this;
        }
        return *this;
    }
    Fleet& operator=(Fleet&& other) {
        std::swap(index, other.index);
        std::swap(side, other.side);
        std::swap(xp, other.xp);
        std::swap(score, other.score);
        std::swap(value, other.value);
        std::swap(race, other.race);
        std::swap(militaryCenters, other.militaryCenters);
        std::swap(salvage, other.salvage);
        std::swap(roidsAllowedToCap, other.roidsAllowedToCap);
        std::swap(battleFormations, other.battleFormations);
        for (BattleFormation& bf : battleFormations) {
            bf.fleet = this;
        }
        return *this;
    }

    Maybe<const BattleFormation> stolen(const std::string& shipTypeName) const;
    Maybe<BattleFormation> stolen(const std::string& shipTypeName);
    void addStolen(const ShipType& shipType, NumberOfShips atStartOfBattle);

    void addXP(XP xp);
    void grantSalvage(Cost salvage);
    Cost fleetCost() const;
    Cost eligibleForSalvage() const;
    Cost costStolen() const;
    Cost costLost() const;
    Cost costChange() const;

    Maybe<const BattleFormation> battleFormation(const ShipType& shipType) const;
    Maybe<BattleFormation> battleFormation(const ShipType& shipType);
    Maybe<const BattleFormation> battleFormation(const std::string& shipName) const;
    Maybe<BattleFormation> battleFormation(const std::string& shipName);
    void addBattleFormation(const ShipType& shipType, NumberOfShips atStartOfBattle);

    Maybe<const BattleFormation> firstPod() const;
    std::vector<const BattleFormation*> combatShips(Class clazz) const;
    std::vector<const BattleFormation*> allAnti(Class clazz) const;

    int64_t index;
    Side side;
    XP xp = 0;
    Score score = 0;
    Value value = 0;
    Race race = Race::None;
    NumberOfShips militaryCenters = 0;
    Cost salvage = 0;
    NumberOfShips roidsAllowedToCap = 0;
    Guns gunsAssignedToRoids = 0;
    std::vector<BattleFormation> battleFormations;
    std::vector<BattleFormation> stolenBattleFormations;
};
bool operator==(const Fleet& lhs, const Fleet& rhs);
bool operator!=(const Fleet& lhs, const Fleet& rhs);

#endif /* FLEET_H */
