// This file is part of mzCalc.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License version 2, as published by the
// Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// mzCalc. If not, see <http://www.gnu.org/licenses/>.
//
// This work is Copyright (C) 2016 by Josse van der Plaat
//
// Individual portions may be copyright by individual contributors, and are
// included in this collective work with permission of the copyright owners.

#ifndef MAYBE_H
#define MAYBE_H

#include <functional>

template<typename T>
struct Maybe {
public:
    explicit Maybe(T* t = nullptr)
        : m_t(t ? t : nullptr)
        , m_owner(false)
    { }
    explicit Maybe(T* t, bool ok)
        : m_t(ok ? t : nullptr)
        , m_owner(false)
    { }
    ~Maybe() noexcept {
        if (m_owner) {
            delete m_t;
        }
    }
    T& operator*() const noexcept {
        return *m_t;
    }
    T* operator->() const noexcept {
        return m_t;
    }
    explicit operator bool() const noexcept {
        return m_t;
    }
    explicit operator T() const noexcept {
        return *m_t;
    }

    template<typename Result, typename MemberFunctionPointer, typename... Args>
    Maybe<Result> exec(MemberFunctionPointer memberFuncPtr, Args&&... args) {
        if (m_t) {
            return Maybe<Result>(new Result((m_t->*memberFuncPtr)(std::forward<Args>(args)...)),
                                 /* Valid */ true,
                                 /* Owner */ true);
        }
        else {
            return Maybe<Result>();
        }
    }

    template<typename Result, typename MemberFunctionPointer, typename... Args>
    const Maybe<Result> exec(MemberFunctionPointer const memberFuncPtr, Args&&... args) const {
        if (m_t) {
            return Maybe<Result>(new Result((m_t->*memberFuncPtr)(std::forward<Args>(args)...)),
                                 /* Valid */ true,
                                 /* Owner */ true);
        }
        else {
            return Maybe<Result>();
        }
    }

    template<typename Result, typename MemberPointer>
    Maybe<Result> get(MemberPointer memberPtr) noexcept {
        if (m_t) {
            return Maybe<Result>(new Result(m_t->*memberPtr),
                                 /* Valid */ true,
                                 /* Owner */ true);
        }
        else {
            return Maybe<Result>();
        }
    }
    template<typename Result, typename MemberPointer>
    const Maybe<Result> get(MemberPointer const memberPtr) const noexcept {
        if (m_t) {
            return Maybe<Result>(new Result(m_t->*memberPtr),
                                 /* Valid */ true,
                                 /* Owner */ true);
        }
        else {
            return Maybe<Result>();
        }
    }


    template<typename AnyOtherT> friend class Maybe;
private:
    explicit Maybe(T* t, bool ok, bool owner)
        : m_t(ok ? t : nullptr)
        , m_owner(owner)
    { }

private:
    T* m_t;

    // If true, m_t is deleted when this is deleted.
    bool m_owner;
};

#endif /* MAYBE_H */
