// This file is part of mzCalc.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License version 2, as published by the
// Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// mzCalc. If not, see <http://www.gnu.org/licenses/>.
//
// This work is Copyright (C) 2016 by Josse van der Plaat
//
// Individual portions may be copyright by individual contributors, and are
// included in this collective work with permission of the copyright owners.

#include "battle.h"

#include "logger.h"

#include <sstream>
#include <numeric>

#include <cmath>

Maybe<const Fleet> Battle::fleet(const Side& side, int64_t index) const {
    auto iter = std::find_if(fleets.cbegin(), fleets.cend(),
                             [&side, index](const Fleet& f) -> bool {
                                 return f.side == side && f.index == index;
                             });
    return Maybe<const Fleet>(&*iter, iter != fleets.end());
}

Maybe<Fleet> Battle::fleet(const Side& side, int64_t index) {
    auto iter = std::find_if(fleets.begin(), fleets.end(),
                             [&side, index](const Fleet& f) -> bool {
                                 return f.side == side && f.index == index;
                             });
    return Maybe<Fleet>(&*iter, iter != fleets.end());
}

Maybe<Fleet> Battle::addFleet(const Side& side) {
    int64_t lastOnSide = -1; // If we have no fleets, our first fleet should
    // be 0.
    for (const Fleet& f : fleets) {
        if (f.side == side) {
            ++lastOnSide;
        }
    }
    fleets.emplace_back(lastOnSide + 1, side);
    return Maybe<Fleet>(&fleets.back(), true);
}

////////////////////////////////////////////////////////////////////////////////

Maybe<const BattleFormation> Battle::battleFormation(const Side& side, int64_t index, const ShipType& type) const {
    return battleFormation(side, index, type.name.c_str());
}

Maybe<BattleFormation> Battle::battleFormation(const Side& side, int64_t index, const ShipType& type) {
    return battleFormation(side, index, type.name.c_str());
}

Maybe<const BattleFormation> Battle::battleFormation(const Side& side, int64_t index, const char* name) const {
    Maybe<const Fleet> f = fleet(side, index);
    if (f) {
        return f->battleFormation(name);
    }
    else {
        return Maybe<const BattleFormation>(nullptr);
    }
}

Maybe<BattleFormation> Battle::battleFormation(const Side& side, int64_t index, const char* name) {
    Maybe<Fleet> f = fleet(side, index);
    if (f) {
        return f->battleFormation(name);
    }
    else {
        return Maybe<BattleFormation>(nullptr);
    }
}

bool Battle::addBattleFormation(const Side& side, int64_t index, const ShipType& type, NumberOfShips present) {
    Maybe<Fleet> f = fleet(side, index);
    if (f) {
        f->addBattleFormation(type, present);
    }
    return bool(f);
}

////////////////////////////////////////////////////////////////////////////////

Cost Battle::fleetCost(Side side) const {
    return std::accumulate(fleets.cbegin(),
                           fleets.cend(),
                           Cost(0),
                           [side](Cost change, const Fleet& f) -> Cost {
                               return change + (f.side == side
                                                ? f.fleetCost()
                                                : 0);
                           });
}

Cost Battle::eligibleForSalvage(Side side) const {
    return std::accumulate(fleets.cbegin(),
                           fleets.cend(),
                           Cost(0),
                           [side](Cost change, const Fleet& f) -> Cost {
                               return change + (f.side == side
                                                ? f.eligibleForSalvage()
                                                : 0);
                           });
}

Cost Battle::costStolen(Side side) const {
    return std::accumulate(fleets.cbegin(),
                           fleets.cend(),
                           Cost(0),
                           [side](Cost change, const Fleet& f) -> Cost {
                               return change + (f.side == side
                                                ? f.costStolen()
                                                : 0);
                           });
}

Cost Battle::costLost(Side side) const {
    return std::accumulate(fleets.cbegin(),
                           fleets.cend(),
                           Cost(0),
                           [side](Cost change, const Fleet& f) -> Cost {
                               return change + (f.side == side
                                                ? f.costLost()
                                                : 0);
                           });
}

Cost Battle::salvage(Side side) const {
    return std::accumulate(fleets.cbegin(),
                           fleets.cend(),
                           Cost(0),
                           [side](Cost change, const Fleet& f) -> Cost {
                               return change + (f.side == side
                                                ? f.salvage
                                                : 0);
                           });
}

Cost Battle::costChange(Side side) const {
    return costStolen(side) + salvage(side) - costLost(side);
}

////////////////////////////////////////////////////////////////////////////////

bool Battle::landableFor(Side side, Verbosity verbosity) const {
    bool landable = true;
    Logger log(verbosity);
    if (side == Side::Attack) {
        Maybe<const BattleFormation> roids = battleFormation(Side::Defense, 0, "Asteroid");
        if (roids) {
            Cost attackerChange = costChange(Side::Attack);
            // For very low losses, we land if we capped any roids at all.
            if (attackerChange >= -1000 && roids->capped() > 0) {
                landable = true;
                log.trace("Attacker change: %+ld = %s\n",
                          attackerChange,
                          landable ? "land" : "recall");
            }
            else {
                Cost defenderChange = costChange(Side::Defense);
                float ratio = float(defenderChange) / attackerChange;
                // An attack is landable if either we repay fairly quickly, or
                // defender loses more than us. The more the defender loses, the
                // longer a repay period we accept.
                float maxRepayTicks =
                    +  3.7*std::pow(ratio - 0.4, 3)
                    -  5.0*std::pow(ratio - 0.4, 2)
                    +  4.0*        (ratio - 0.4)
                    + 14.0;
                float repayTicks = roids->capped() > 0
                    ? float(-attackerChange) / (roids->capped() * 250 * (1.0 + 0.25 + 0.24 + 0.15)) // Assume income bonus from Corp + 30 FCs
                    : attackerChange / 10; // Without roid gain, we assume we gain 10 value per tick.
                landable = repayTicks < maxRepayTicks;
                log.trace("Attacker change: %+ld, Defender change: %+ld, Ratio: %.2f, Repay ticks: %f, Max repay ticks: %f = %s\n",
                          attackerChange,
                          defenderChange,
                          ratio,
                          repayTicks,
                          maxRepayTicks,
                          landable ? "land" : "recall");
            }
        }
        else {
            // Attacks on roidless planets are never landable.
            landable = false;
            log.trace("No roids = recall\n");
        }
    }
    else {
        // TODO: Not implemented.
        log.error("Battle::landableFor not implemented for defenders!\n");
    }
    return landable;
}
bool Battle::betterThan(const Battle& other, Side forSide, Verbosity verbosity) const {
    Logger log(verbosity);
    bool better = false;
    if (forSide == Side::Attack) {
        // TODO: Not implemented.
        log.error("Battle::betterThan not implemented for attackers!\n");
    }
    else {
        bool otherLandable = other.landableFor(!forSide, verbosity);
        bool landable      =       landableFor(!forSide, verbosity);
        if (otherLandable != landable) {
            better = !otherLandable && landable;
        }
        else {
            Maybe<const BattleFormation> otherRoids = other.battleFormation(Side::Defense, 0, "Asteroid");
            Maybe<const BattleFormation>      roids =       battleFormation(Side::Defense, 0, "Asteroid");
            if (otherRoids && roids) {
                float otherRepay = std::ceil(float(-other.costChange(Side::Attack)) / (otherRoids->capped() * 250 * (1.0 + 0.25 + 0.24 + 0.3)));
                float      repay = std::ceil(float(-      costChange(Side::Attack)) / (     roids->capped() * 250 * (1.0 + 0.25 + 0.24 + 0.3)));
                better = otherRepay > repay;
            }
            else {
                log.error("Missing home planet roids in one of the two battles!\n");
            }
        }
    }
    return better;
}

////////////////////////////////////////////////////////////////////////////////

std::string Battle::toString() const {
    std::stringstream ss;
    std::string str(250, '\0');

    ss << "|------------------------------------------------------------------|\n"
          "| Fleet | Ship       |  Start | Hugged |   Lost | Stolen |   Value |\n";
    Side side(Side::Defense);
    int64_t index = -1;
    for (const Fleet& fleet : fleets) {
        if (fleet.side != side) {
            side = fleet.side;
            index = fleet.index;
            ss << "|------------------------------------------------------------------|\n";
        }
        else if (fleet.index != index) {
            ss << "|------------------------------------------------------------------|\n";
        }
        for (const BattleFormation& bf : fleet.battleFormations) {
            int written = snprintf(&str[0],
                                   str.size() - 1,
                                   "| %s-%ld | %-10s | %6ld | %6ld | %6ld | %6ld | %+7ld |\n",
                                   bf.fleet->side.toString().substr(0, 3).c_str(),
                                   bf.fleet->index,
                                   bf.shipType.name.substr(0, 9).c_str(),
                                   bf.atStartOfBattle,
                                   bf.emped(),
                                   bf.killed() + bf.abandoned() + bf.stolen() + bf.capped(),
                                   int64_t(0),
                                   -(bf.killed() + bf.stolen() + bf.capped() + bf.structureKilled() + bf.abandoned()) * bf.shipType.cost / 100);
            str[written + 1] = '\0';
            ss << str.c_str();
        }
        for (const BattleFormation& bf : fleet.stolenBattleFormations) {
            int written = snprintf(&str[0],
                                   str.size() - 1,
                                   "| %s-%ld | %-10s |      0 |      0 |      0 | %6ld | %+7ld |\n",
                                   bf.fleet->side.toString().substr(0, 3).c_str(),
                                   bf.fleet->index,
                                   bf.shipType.name.substr(0, 9).c_str(),
                                   bf.atStartOfBattle,
                                   bf.atStartOfBattle * bf.shipType.cost / 100);
            str[written + 1] = '\0';
            ss << str.c_str();
        }
    }
    ss << "|------------------------------------------------------------------|";

    return ss.str();
}

bool operator==(const Battle& lhs, const Battle& rhs) {
    return lhs.fleets.size() == rhs.fleets.size() &&
        std::mismatch(lhs.fleets.cbegin(), lhs.fleets.cend(),
                      rhs.fleets.cbegin(), rhs.fleets.cend()).first == lhs.fleets.cend();
}

bool operator!=(const Battle& lhs, const Battle& rhs) {
    return !(lhs == rhs);
}
