// This file is part of mzCalc.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License version 2, as published by the
// Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// mzCalc. If not, see <http://www.gnu.org/licenses/>.
//
// This work is Copyright (C) 2016 by Josse van der Plaat
//
// Individual portions may be copyright by individual contributors, and are
// included in this collective work with permission of the copyright owners.

#include "bcalc.h"

#include "battle.h"
#include "fleet.h"
#include "shiptype.h"

#include "logger.h"

#include <vector>
#include <limits>
#include <map>
#include <numeric>

#include <cmath>

namespace {
struct ClassSideType {
    Class clazz;
    Side side;
    GunType type;

    ClassSideType(Class clazz, Side side, GunType type)
        : clazz(clazz)
        , side(side)
        , type(type)
    { }
};
bool operator<(const ClassSideType& lhs, const ClassSideType& rhs) {
    return lhs.clazz == rhs.clazz
        ? (lhs.side == rhs.side
            ? lhs.type < rhs.type
            : lhs.side < rhs.side)
        : lhs.clazz < rhs.clazz;
}

struct DamageDealt {
    Guns damageDealt = 0;
    Guns overkill = 0;
    std::vector<BattleFormation*> contributing;
    std::vector<BattleFormation> stolen;
};
}

void bcalc(Battle& battle,
           bool useRepeatSameTargetIndex,
           Verbosity verbosity) {
    Logger log(verbosity);

    std::vector<BattleFormation*> battleFormations;
    for (Fleet& fleet : battle.fleets) {
        for (BattleFormation& bf : fleet.battleFormations) {
            battleFormations.emplace_back(&bf);
        }
    }

    // Step 1.5: Assign potential roid cap to each fleet.
    Value totalAttackValue = std::accumulate(battle.fleets.begin(),
                                             battle.fleets.end(),
                                             Value(0),
                                             [](Value sum, const Fleet& fleet) -> Value {
                                                 return sum + (fleet.side == Side::Attack
                                                               ? fleet.fleetCost()
                                                               : 0);
                                             });
    Value homeValue = -1;
    NumberOfShips totalCap = -1;
    if (Maybe<Fleet> home = battle.fleet(Side::Defense, 0)) {
        homeValue = home->value;
        if (Maybe<BattleFormation> asteroids = home->battleFormation("Asteroid")) {
            totalCap = asteroids->atStartOfBattle * 0.25f;
        }
        else {
            log.error("Missing home roids, cannot calculate roid capture.\n");
        }
    }
    else {
        log.error("Missing home fleet, cannot calculate roid capture.\n");
    }
    if (totalCap > 0 && totalAttackValue > 0) {
        for (Fleet& fleet : battle.fleets) {
            if (fleet.side == Side::Attack) {
                float bashReduction = 1.0f;
                if (fleet.value > 0 && homeValue > 0) {
                    bashReduction = std::min(1.0f, std::sqrt(float(homeValue) / fleet.value));
                }
                else {
                    log.info("Missing planet value for fleet %s-%ld, assuming max cap of 25%%.\n",
                              fleet.side.toString().substr(0, 3).c_str(),
                              fleet.index);
                }
                fleet.roidsAllowedToCap = totalCap * fleet.fleetCost() / totalAttackValue * bashReduction;
            }
        }
    }
    else if (totalCap < 0 || totalAttackValue < 0) {
        log.error("No home planet roid count and/or attacking fleets found, cannot calculate roid capture.\n");
    }

    // Step 2: find Inits with BattleFormations in them, in ascending order.
    std::vector<Init> initsPresent(battleFormations.size(), 0);
    std::transform(battleFormations.begin(), battleFormations.end(),
                   initsPresent.begin(),
                   [](const BattleFormation* bf) -> Init {
                       return bf->shipType.init;
                   });
    std::sort(initsPresent.begin(), initsPresent.end());
    auto last = std::unique(initsPresent.begin(), initsPresent.end());
    initsPresent.erase(last, initsPresent.end());

    // Step 3: iterate through Inits.
    for (Init init : initsPresent) {
        log.trace("Init %ld\n", init);
        for (std::size_t targetIndex = 0; targetIndex < MAX_TARGETS; ++targetIndex) {
            Eff eff = effs[targetIndex];
            log.trace("T%ld @ %ld%%\n", targetIndex+1, eff);
            // As long as we overkill a given target class, repeat the same
            // target class.
            bool repeatSameTargetIndex;
            do {
                repeatSameTargetIndex = false;
                std::map<ClassSideType, DamageDealt> firedAt;
                for (BattleFormation* bf : battleFormations) {
                    if (bf->shipType.init == init) {
                        NumberOfShips gunsReady = bf->gunsReady();
                        Class target = bf->shipType.target(targetIndex);
                        if (gunsReady > 0) {
                            if (target == Class::Asteroid) {
                                bf->fleet->gunsAssignedToRoids += gunsReady;
                            }
                            if (target != Class::None) {
                                log.trace("Shoot:    (%4ld - %4ldE - %4ldK - %4ldS) %-4s (side=%s, fleet=%ld, target=%2s,  guns=%-3ld) firing %5ld/%5ld guns\n",
                                          bf->atStartOfBattle,
                                          bf->emped(),
                                          bf->killed(),
                                          bf->stolen(),
                                          bf->shipType.name.substr(0, 4).c_str(),
                                          bf->fleet->side.toString().substr(0, 3).c_str(),
                                          bf->fleet->index,
                                          target.toString().substr(0, 2).c_str(),
                                          bf->shipType.guns,
                                          gunsReady,
                                          bf->atStartOfBattle * bf->shipType.guns);
                                // Firing guns.
                                DamageDealt& dd = firedAt[ClassSideType(target,
                                                                     !bf->fleet->side,
                                                                     bf->shipType.type)];
                                dd.damageDealt += gunsReady;
                                dd.contributing.push_back(bf);
                            }
                        }
                    }
                }
                // We need to store how many guns each BattleFormation has ready
                // at the start of this init/T/repeat.
                std::map<BattleFormation*, Guns> gunsAtStartOfInit;
                // Step 4: applying damage.
                for (auto& pair : firedAt) {
                    ClassSideType cs = pair.first;
                    DamageDealt&  dd = pair.second;
                    log.trace("%sing %ss were dealt %ld %s damage\n",
                              cs.side.toString().c_str(),
                              cs.clazz.toString().c_str(),
                              dd.damageDealt,
                              cs.type.toString().c_str());
                    // Figure out which BattleFormations were targetted and how
                    // many ships are in them in total.
                    std::vector<BattleFormation*> targetted;
                    NumberOfShips totalTargets = 0;
                    for (BattleFormation* bf : battleFormations) {
                        if (bf->fleet->side == cs.side && bf->shipType.clazz == cs.clazz) {
                            targetted.push_back(bf);
                            gunsAtStartOfInit[bf] = bf->gunsReady();
                            switch (cs.type.type) {
                            case GunType::None: {
                                log.trace("None: Nothing to do\n");
                            } break;
                            case GunType::Kill:
                            case GunType::Steal: {
                                totalTargets += bf->alive();
                            } break;
                            case GunType::EMP: {
                                totalTargets += bf->unemped();
                            } break;
                            case GunType::Pod: {
                                totalTargets += bf->uncapped();
                            } break;
                            case GunType::StructureKill: {
                                totalTargets += bf->unstructureKilled();
                            } break;
                            }
                        }
                    }
                    if (totalTargets <= 0) {
                        // Nothing to shoot at.
                        log.trace("Miss:     (     -     E -     K -     S)      (side=%s, fleet= , target=%2s, armor=   ) +   0K, %5ld overkill\n",
                                  cs.side.toString().substr(0, 3).c_str(),
                                  cs.clazz.toString().substr(0, 2).c_str(),
                                  dd.damageDealt);
                        dd.overkill += dd.damageDealt;
                    }
                    else {
                        // Now we can determine how many guns fire at each BattleFormation.
                        for (BattleFormation* bf : targetted) {
                            switch (cs.type.type) {
                            case GunType::None: {
                                log.trace("None: Nothing to do\n");
                            } break;
                            case GunType::Kill: {
                                log.trace("Kill:     (%4ld - %4ldE - %4ldK - %4ldS) %4s (side=%s, fleet=%ld,  class=%2s, armor=%-3ld)",
                                          bf->atStartOfBattle,
                                          bf->emped(),
                                          bf->killed(),
                                          bf->stolen(),
                                          bf->shipType.name.substr(0, 4).c_str(),
                                          bf->fleet->side.toString().substr(0, 3).c_str(),
                                          bf->fleet->index,
                                          bf->shipType.clazz.toString().substr(0, 2).c_str(),
                                          bf->shipType.armor);
                                Guns guns = dd.damageDealt * bf->alive() / totalTargets;
                                NumberOfShips canKill = guns * eff / 100 / bf->shipType.armor;
                                NumberOfShips killed = bf->kill(canKill);
                                // If we had more guns left over than
                                // because of killing part of the last ship,
                                // we should shoot again at the current
                                // target class. Otherwise, it's safe to
                                // move to the next.
                                if (canKill > killed) {
                                    repeatSameTargetIndex = true;
                                }
                                Guns overkill = guns - std::ceil(killed * bf->shipType.armor * 100 / eff);
                                dd.overkill += overkill;
                                log.trace(" +%4ldK, %5ld overkill\n",
                                          killed,
                                          overkill);
                            } break;
                            case GunType::EMP: {
                                log.trace("EMP:      (%4ld - %4ldE - %4ldK - %4ldS) %4s (side=%s, fleet=%ld,  class=%2s,  eres=%-3ld)",
                                          bf->atStartOfBattle,
                                          bf->emped(),
                                          bf->killed(),
                                          bf->stolen(),
                                          bf->shipType.name.substr(0, 4).c_str(),
                                          bf->fleet->side.toString().substr(0, 3).c_str(),
                                          bf->fleet->index,
                                          bf->shipType.clazz.toString().substr(0, 2).c_str(),
                                          bf->shipType.eres);
                                if (totalTargets > 0) {
                                    float empEff = 1.0f - bf->shipType.eres / 100.0f;
                                    // Here, rounding error down causes the engine
                                    // to believe one fewer ship can be emped than
                                    // really can be.
                                    Guns guns = dd.damageDealt * bf->unemped() / totalTargets;
                                    NumberOfShips canEmp = guns * eff * std::nextafter(empEff, 1.0f) / 100;
                                    NumberOfShips emped = bf->emp(canEmp);
                                    if (canEmp > emped) {
                                        repeatSameTargetIndex = true;
                                    }
                                    // Here, rounding error up causes the engine to
                                    // believe one more gun is required to emp all
                                    // the ships.
                                    Guns overkill = std::ceil(guns - emped * 100 / std::nextafter(empEff, 0.0f) / eff);
                                    dd.overkill += overkill;
                                    log.trace(" +%4ldE, %5ld overkill\n",
                                              emped,
                                              overkill);
                                }
                                else {
                                    // All targets already emped.
                                    dd.overkill += dd.damageDealt;
                                    log.trace(" +    , %5ld overkill\n",
                                              dd.overkill);
                                }
                            } break;
                            case GunType::Steal: {
                                log.trace("Steal:    (%4ld - %4ldE - %4ldK - %4ldS) %4s (side=%s, fleet=%ld,  class=%2s, armor=%-3ld)",
                                          bf->atStartOfBattle,
                                          bf->emped(),
                                          bf->killed(),
                                          bf->stolen(),
                                          bf->shipType.name.substr(0, 4).c_str(),
                                          bf->fleet->side.toString().substr(0, 3).c_str(),
                                          bf->fleet->index,
                                          bf->shipType.clazz.toString().substr(0, 2).c_str(),
                                          bf->shipType.armor);
                                if (totalTargets > 0) {
                                    NumberOfShips canSteal = dd.damageDealt * bf->alive() / totalTargets * eff / bf->shipType.armor / 100;
                                    NumberOfShips stolen = bf->steal(canSteal);
                                    if (canSteal > stolen) {
                                        repeatSameTargetIndex = true;
                                    }
                                    Guns overkill = (canSteal -  stolen) * bf->shipType.armor;
                                    dd.overkill += overkill;
                                    dd.stolen.emplace_back(bf->shipType, stolen, nullptr);
                                    log.trace(" +%4ldS, %5ld overkill\n",
                                              stolen,
                                              overkill);
                                }
                                else {
                                    // All targets already dead.
                                    dd.overkill += dd.damageDealt;
                                    log.trace(" +    , %5ld overkill\n",
                                              dd.overkill);
                                }

                            } break;
                            case GunType::Pod: {
                                // Calcualtion of roid capture is delayed until
                                // after damage application.
                            } break;
                            case GunType::StructureKill: {
                                log.trace("SK:       (%4ld - %4ldSK               ) %4s (side=%s, fleet=%ld,  class=%2s, armor=%-3ld)",
                                          bf->atStartOfBattle,
                                          bf->structureKilled(),
                                          bf->shipType.name.substr(0, 4).c_str(),
                                          bf->fleet->side.toString().substr(0, 3).c_str(),
                                          bf->fleet->index,
                                          bf->shipType.clazz.toString().substr(0, 2).c_str(),
                                          bf->shipType.armor);
                                if (totalTargets > 0) {
                                    Guns guns = dd.damageDealt * bf->alive() / totalTargets;
                                    NumberOfShips canDestroy = guns * eff / 100 / bf->shipType.armor;
                                    NumberOfShips destroyed = bf->structureKill(canDestroy);
                                    if (canDestroy > destroyed) {
                                        repeatSameTargetIndex = true;
                                    }
                                    Guns overkill = guns - std::ceil(destroyed * bf->shipType.armor * 100 / eff);
                                    dd.overkill += overkill;
                                    log.trace(" +%4ldSK, %5ld overkill\n",
                                              destroyed,
                                              overkill);
                                }
                                else {
                                    // All targets already dead.
                                    dd.overkill += dd.damageDealt;
                                    log.trace(" +    , %5ld overkill\n",
                                              dd.overkill);
                                }
                            } break;
                            }
                        }
                    }

                    // Step 5: distribute overkill for T2 onwards, stolen ships
                    // and abandoned ships among firing BattleFormations,
                    // according to the total number of guns present in each
                    // BattleFormation at the start of this init/T/repeat.
                    Guns totalGuns = std::accumulate(dd.contributing.begin(),
                                                     dd.contributing.end(),
                                                     Guns(0),
                                                     [&gunsAtStartOfInit](Guns sum, BattleFormation* bf) -> Guns {
                                                         auto iter = gunsAtStartOfInit.find(bf);
                                                         return sum + (iter == gunsAtStartOfInit.end()
                                                                       ? bf->gunsReady()
                                                                       : iter->second);
                                                     });
                    // // C++17
                    // Guns totalGuns = std::transform_reduce(dd.contributing.begin(),
                    //                                        dd.contributing.end(),
                    //                                        Guns(0),
                    //                                        [&gunsAtStartOfInit](const BattleFormation* bf) -> Guns {
                    //                                            auto iter = gunsAtStartOfInit.find(bf);
                    //                                            return sum + (iter == gunsAtStartOfInit.end()
                    //                                                          ? bf->gunsReady()
                    //                                                          : iter->second);
                    //                                        })
                    for (BattleFormation* bf : dd.contributing) {
                        for (const BattleFormation& s : dd.stolen) {
                            // If bf was targetted at this init, we need the
                            // number of guns at the start of this init. If not,
                            // we can just take the number of guns at the end of
                            // this init.
                            auto iter = gunsAtStartOfInit.find(bf);
                            Guns gunsReady = iter == gunsAtStartOfInit.end()
                                ? bf->gunsReady()
                                : iter->second;
                            // TODO: Rounding may cause a(?) stolen ship to
                            // evaporate.
                            NumberOfShips stolen = s.atStartOfBattle * gunsReady / totalGuns;
                            if (stolen > 0) {
                                bf->fleet->addStolen(s.shipType, stolen);
                                bf->abandon(s.shipType.cost * stolen / bf->shipType.cost);
                            }
                        }
                        // TODO: Abandon SKs by value, probably.
                        auto iter = gunsAtStartOfInit.find(bf);
                        Guns gunsReady = iter == gunsAtStartOfInit.end()
                            ? bf->gunsReady()
                            : iter->second;
                        if (dd.overkill > 0) {
                            // TODO: Rounding may cause a(?) non-firing gun to
                            // evaporate.
                            Guns overkill = dd.overkill * gunsReady / totalGuns;

                            log.trace("Overkill: (%4ld - %4ldE - %4ldK - %4ldS) %4s (side=%s, fleet=%ld) %5ld returned\n",
                                      bf->atStartOfBattle,
                                      bf->emped(),
                                      bf->killed(),
                                      bf->stolen(),
                                      bf->shipType.name.substr(0, 4).c_str(),
                                      bf->fleet->side.toString().substr(0, 3).c_str(),
                                      bf->fleet->index,
                                      overkill);
                            bf->fireGuns(gunsReady - overkill);
                        }
                        else {
                            bf->fireGuns(gunsReady);
                        }
                    }
                }
                // Step 5.5: Roid capture
                if (totalCap > 0) {
                    for (Fleet& fleet : battle.fleets) {
                        if (fleet.side == Side::Attack && fleet.gunsAssignedToRoids > 0) {
                            Maybe<Fleet> home = battle.fleet(Side::Defense, 0);
                            Maybe<BattleFormation> roids = home->battleFormation("Asteroid");
                            log.trace("Pod:      (%4ld - %4ldC                ) %4s (side=%s, fleet=%ld,  class=%2s, armor=%-3ld)",
                                      roids->atStartOfBattle,
                                      roids->capped(),
                                      roids->shipType.name.substr(0, 4).c_str(),
                                      home->side.toString().substr(0, 3).c_str(),
                                      home->index,
                                      roids->shipType.clazz.toString().substr(0, 2).c_str(),
                                      roids->shipType.armor);

                            NumberOfShips canCap = fleet.gunsAssignedToRoids / roids->shipType.armor;
                            NumberOfShips capped = roids->cap(std::min(fleet.roidsAllowedToCap, canCap));
                            if (canCap > capped) {
                                repeatSameTargetIndex = true;
                            }
                            fleet.roidsAllowedToCap -= capped;
                            fleet.gunsAssignedToRoids = 0;
                            fleet.addStolen(roids->shipType, capped);
                            log.trace(" +%4ldC\n",
                                      capped);
                        }
                    }
                }
            } while (useRepeatSameTargetIndex && repeatSameTargetIndex);
        }
        // Step 6 and 7: Iterate through the above T1-T3 loop.

        log.trace("\n");
    }
    // Step 8: Iterate through the above init loop.

    // Step 8.5: Abandon the SKs that fired.
    for (BattleFormation* bf : battleFormations) {
        if (bf->shipType.type == GunType::StructureKill) {
            // The 'lore' is that the ships are abandoned, but the mechanics
            // work out better if we set them to killed.
            bf->kill(bf->gunsFired() / bf->shipType.guns);
        }
    }

    // Step 9: salvage

    Cost attackerSalvage = battle.eligibleForSalvage(Side::Attack);
    Cost totalDefenseCost = battle.fleetCost(Side::Defense);
    if (totalDefenseCost > 0) {
        for (Fleet& fleet : battle.fleets) {
            if (fleet.side == Side::Defense) {
                Cost startCost = fleet.fleetCost();
                // Defense salvage is 40% of the cost of killed and abandoned ships
                // in each defending fleet returned to that fleet.
                fleet.grantSalvage(fleet.eligibleForSalvage() * 0.4f);
                // Attack salvage is 20% of killed and abandoned attacker ships is split among
                // defenders according to their fleet values.
                fleet.grantSalvage(std::nextafter(attackerSalvage * 0.2f, attackerSalvage) * startCost / totalDefenseCost);
                // Salvage is limited to 20% of original fleet cost.
                Cost endCost =
                    std::accumulate(fleet.battleFormations.cbegin(),
                                    fleet.battleFormations.cend(),
                                    Cost(0),
                                    [](Cost cost, const BattleFormation& bf) -> Cost {
                                        // Skip ships that are not really ships:
                                        // Asteroids and Structures.
                                        return cost + (bf.shipType.clazz == Class::Asteroid || bf.shipType.clazz == Class::Structure
                                                       ? 0
                                                       : bf.shipType.cost * bf.alive());
                                    }) +
                    std::accumulate(fleet.stolenBattleFormations.cbegin(),
                                    fleet.stolenBattleFormations.cend(),
                                    Cost(0),
                                    [](Cost cost, const BattleFormation& bf) -> Cost {
                                        // Skip ships that are not really ships:
                                        // Asteroids and Structures.
                                        return cost + (bf.shipType.clazz == Class::Asteroid || bf.shipType.clazz == Class::Structure
                                                       ? 0
                                                       : bf.shipType.cost * bf.atStartOfBattle);
                                    });
                if (endCost + fleet.salvage > startCost * 1.2f) {
                    Cost reducedSalvage = startCost * 6 / 5 - endCost;
                    log.trace("Reducing salvage for fleet going from %ld to %ld cost, from %ld to %ld\n",
                              startCost,
                              endCost,
                              fleet.salvage,
                              reducedSalvage);
                    fleet.salvage = reducedSalvage;
                }
            }
        }
    }

    // Step 10: XP
    if (Maybe<Fleet> home = battle.fleet(Side::Defense, 0)) {
        for (Fleet& fleet : battle.fleets) {
            if (fleet.side == Side::Attack) {
                if (Maybe<BattleFormation> roids = fleet.stolen("Asteroid")) {
                    float bravery = 0.0f;
                    if (fleet.score > 0 && fleet.value > 0) {
                        bravery = std::max(0.2f, std::min(2.2f, float(home->score) / fleet.score) - 0.2f) *
                                  std::max(0.2f, std::min(1.8f, float(home->value) / fleet.value) - 0.1f) /
                          (0.6f + std::max(4.0f,                float(fleet.score) / fleet.value) * 0.1f);
                    }
                    XP xp = roids->atStartOfBattle * 10 * bravery * (1 + float(fleet.militaryCenters) / 200);
                    fleet.addXP(xp);
                    log.trace("Fleet %s-%ld capped %ld roids with %f bravery for %ld XP\n",
                              fleet.side.toString().substr(0, 3).c_str(),
                              fleet.index,
                              roids->atStartOfBattle,
                              bravery,
                              xp);
                }
            }
        }
    }
}
