// This file is part of mzCalc.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License version 2, as published by the
// Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// mzCalc. If not, see <http://www.gnu.org/licenses/>.
//
// This work is Copyright (C) 2016 by Josse van der Plaat
//
// Individual portions may be copyright by individual contributors, and are
// included in this collective work with permission of the copyright owners.

#include "fillbattle.h"

#include "battle.h"
#include "fleet.h"
#include "stats.h"

#include "logger.h"
#include "maybe.h"

inline bool Tokenizer::hasNext() const {
    return index < tokens.size();
}
const std::string& Tokenizer::next() {
    if (hasNext()) {
        return tokens[index++];
    }
    else {
        throw TokenizerException("Ran out of tokens!");
    }
}

////////////////////////////////////////////////////////////////////////////////

bool fillBattle(Battle& battle, const Stats& stats, const std::vector<std::string> tokens, Verbosity verbosity) {
    Logger log(verbosity);
    Side::S side = Side::Attack;
    int64_t fleetIndex = -1;
    bool ok = true;
    Tokenizer tokenizer(tokens);
    // Accepted grammar:
    // <battle> := <fleet...> "vs" <fleet...>
    // <fleet>  := "fleet" <value> <score> <ship...>
    // <ship>   := <number> <shipname>
    while (tokenizer.hasNext() && ok) {
        const std::string& token = tokenizer.next();
        try {
            if (token == "fleet" || token == "Fleet") {
                Maybe<Fleet> fleet = battle.addFleet(side);
                fleetIndex = fleet->index;
                const std::string& valueToken = tokenizer.next();
                try {
                    fleet->value = std::stoull(valueToken);
                    const std::string& scoreToken = tokenizer.next();
                    try {
                        fleet->score = std::stoull(scoreToken);
                    }
                    catch (std::invalid_argument ia) {
                        log.error("Expected planet score for fleet %s-%ld, got %s, aborting\n",
                                  Side(side).toString().substr(0, 3).c_str(),
                                  fleetIndex,
                                  scoreToken.c_str());
                        ok = false;
                    }
                }
                catch (std::invalid_argument ia) {
                    log.error("Expected planet value for fleet %s-%ld, got %s, aborting\n",
                              Side(side).toString().substr(0, 3).c_str(),
                              fleetIndex,
                              valueToken.c_str());
                    ok = false;
                }
                log.trace("Added %s fleet with %ld value and %ld score\n",
                          Side(side).toString().c_str(),
                          fleet->value,
                          fleet->score);
            }
            else if (token == "vs" || token == "vs.") {
                if (side == Side::Attack) {
                    log.trace("Switched to adding defense fleets\n");
                    fleetIndex = -1;
                    side = Side::Defense;
                }
                else {
                    log.error("Expected just one 'vs' or 'vs.' token, got a second one, aborting\n");
                    ok = false;
                }
            }
            else if (fleetIndex < 0) {
                log.error("Expected 'fleet' or 'Fleet', got %s, aborting\n",
                          token.c_str());
                ok = false;
            }
            else {
                try {
                    NumberOfShips nos = std::stoull(token);
                    const std::string& shipTypeToken = tokenizer.next();
                    if (Maybe<const ShipType> shipType = stats[shipTypeToken]) {
                        log.trace("Added %ld %s to fleet %s-%ld\n",
                                  nos,
                                  shipType->name.c_str(),
                                  Side(side).toString().c_str(),
                                  fleetIndex);
                        battle.addBattleFormation(side, fleetIndex, *shipType, nos);
                    }
                    else {
                        log.error("Expected ship name, got %s, aborting\n",
                                  shipTypeToken.c_str());
                        ok = false;
                    }
                }
                catch (std::invalid_argument ia) {
                    log.error("Expected number of ships, got %s, aborting\n",
                              token.c_str());
                    ok = false;
                }
            }
        }
        catch (TokenizerException te) {
            log.error("Expected more tokens after %s, but found none, aborting\n",
                      token.c_str());
            ok = false;
        }
    }
    return ok;
}
