// This file is part of mzCalc.
//
// This program is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License version 2, as published by the
// Free Software Foundation.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
// details.
//
// You should have received a copy of the GNU General Public License along with
// mzCalc. If not, see <http://www.gnu.org/licenses/>.
//
// This work is Copyright (C) 2016 by Josse van der Plaat
//
// Individual portions may be copyright by individual contributors, and are
// included in this collective work with permission of the copyright owners.

#include "test.h"

#include "bcalc.h"
#include "stats.h"
#include "analyzestats.h"

#include "battle.h"

#include "logger.h"

#include <map>
#include <string>
#include <cmath>

Stats stats;

float valueRatio = 1.0f;
// bool withRepeatSameTargetIndex = true;
bool withoutRepeatSameTargetIndex = false;
bool withFlak    = true;
bool withoutFlak = false;

void emptyBattle() {
    Battle actual(0, 0);

    Battle expect(actual);

    bcalc(actual, false, Verbosity::None);
    EXPECT_EQUAL(actual, expect);
}

void emptyFleet() {
    Battle actual(0, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"], 0));

    Battle expect(actual);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void emptyFleets() {
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"], 0));

    Battle expect(actual);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void nonEmptyAttack() {
    Battle actual(2, 0);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Harpy"], 5000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  1, *stats["Harpy"], 6000));

    Battle expect(actual);

    bcalc(actual, false, Verbosity::None);
    EXPECT_EQUAL(actual, expect);
}

void noInteraction() {
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Harpy"], 1000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Harpy"], 2000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"], 0));

    Battle expect(actual);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}


// KILL ////////////////////////////////////////////////////////////////////////

void zeroLossKill() {
    // Phantom do not fire at Pegasus. Pegasus kill some Phantom.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Phantom"], 11000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],    0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Pegasus"],   500));

    Battle expect(actual);
    NumberOfShips deadPhantom = 500 * stats["Pegasus"]->guns / stats["Phantom"]->armor;
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Phantom")->kill(deadPhantom), deadPhantom);
    Cost salvage = deadPhantom * stats["Phantom"]->cost * 0.2;
    EXPECT_TRUE(salvage > expect.fleet(Side::Defense, 0)->fleetCost() / 5);
    expect.fleet(Side::Defense, 0)->grantSalvage(500 * stats["Pegasus"]->cost / 5);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void distributeKill() {
    // Phantom and Corsair do not fire at Pegasus and Spectre. Pegasus and
    // Spectre both fire at Phantom and Corsair.
    Battle actual(2, 2);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Phantom"], 20000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  1, *stats["Corsair"], 10000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],    0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Pegasus"],   300));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 1, *stats["Spectre"],  1200));

    Battle expect(actual);
    NumberOfShips deadPhantom =
        300 * 20000 / (20000+10000) * stats["Pegasus"]->guns / stats["Phantom"]->armor +
        1200 * 20000 / (20000+10000) * stats["Spectre"]->guns / stats["Phantom"]->armor;
    NumberOfShips deadCorsair =
        300 * 10000 / (20000+10000) * stats["Pegasus"]->guns / stats["Corsair"]->armor +
        1200 * 10000 / (20000+10000) * stats["Spectre"]->guns / stats["Corsair"]->armor;
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Phantom")->kill(deadPhantom), deadPhantom);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 1)->battleFormation("Corsair")->kill(deadCorsair), deadCorsair);
    Value pegasusCost =  300 * stats["Pegasus"]->cost;
    Value spectreCost = 1200 * stats["Spectre"]->cost;

    Cost salvage0 = (deadCorsair * stats["Corsair"]->cost +
                     deadPhantom * stats["Phantom"]->cost) * 0.2 * pegasusCost / (pegasusCost + spectreCost);
    Cost salvage1 = (deadCorsair * stats["Corsair"]->cost +
                     deadPhantom * stats["Phantom"]->cost) * 0.2 * spectreCost / (pegasusCost + spectreCost);
    Cost salvageCap0 = expect.fleet(Side::Defense, 0)->fleetCost() / 5;
    Cost salvageCap1 = expect.fleet(Side::Defense, 1)->fleetCost() / 5;
    EXPECT_TRUE(salvage0 > salvageCap0);
    EXPECT_TRUE(salvage1 > salvageCap0);

    expect.fleet(Side::Defense, 0)->grantSalvage(salvageCap0);
    expect.fleet(Side::Defense, 1)->grantSalvage(salvageCap1);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void dontTargetDeadShips() {
    // Phantom and Corsair do not fire at Pegasus and Spectre. Pegasus and
    // Spectre both fire at Phantom and Corsair.
    Battle actual(2, 2);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Phantom"], 10000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  1, *stats["Corsair"], 10000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],    0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Pegasus"],  1000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 1, *stats["Clipper"],    10));

    Battle expect(actual);
    NumberOfShips deadPhantom = 10000 * 1000 / (10000+10000) * stats["Pegasus"]->guns / stats["Phantom"]->armor;
    EXPECT_EQUAL(deadPhantom, NumberOfShips(10000));
    NumberOfShips deadCorsair0 = 10000 * 1000 / (10000+10000) * stats["Pegasus"]->guns / stats["Corsair"]->armor;
    EXPECT_TRUE(deadCorsair0 < 10000);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Phantom")->kill(deadPhantom), deadPhantom);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 1)->battleFormation("Corsair")->kill(deadCorsair0), deadCorsair0);

    NumberOfShips deadCorsair1 = 10 * stats["Clipper"]->guns / stats["Corsair"]->armor;
    EXPECT_EQUAL(expect.fleet(Side::Attack, 1)->battleFormation("Corsair")->kill(deadCorsair1), deadCorsair1);

    Value pegasusCost = 1000 * stats["Pegasus"]->cost;
    Value clipperCost =   10 * stats["Clipper"]->cost;

    Cost salvage0 = (deadCorsair0 * stats["Corsair"]->cost +
                     deadCorsair1 * stats["Corsair"]->cost +
                     deadPhantom  * stats["Phantom"]->cost) * 0.2 * pegasusCost / (pegasusCost + clipperCost);
    Cost salvage1 = (deadCorsair0 * stats["Corsair"]->cost +
                     deadCorsair1 * stats["Corsair"]->cost +
                     deadPhantom  * stats["Phantom"]->cost) * 0.2 * clipperCost / (pegasusCost + clipperCost);
    Cost salvageCap0 = expect.fleet(Side::Defense, 0)->fleetCost() / 5;
    Cost salvageCap1 = expect.fleet(Side::Defense, 1)->fleetCost() / 5;
    EXPECT_TRUE(salvage0 > salvageCap0);
    EXPECT_TRUE(salvage1 > salvageCap1);
    expect.fleet(Side::Defense, 0)->grantSalvage(salvageCap0);
    expect.fleet(Side::Defense, 1)->grantSalvage(salvageCap1);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}


void exchangeKill() {
    // Harpy fire at Pegasus. Remaining Pegasus fire at Harpy.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Harpy"],  6700));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Pegasus"], 837));

    Battle expect(actual);
    NumberOfShips deadPegasus = 6700                * stats["Harpy"]->guns   / stats["Pegasus"]->armor;
    NumberOfShips deadHarpy   = (837 - deadPegasus) * stats["Pegasus"]->guns / stats["Harpy"]->armor;
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Pegasus")->kill(deadPegasus), deadPegasus);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Harpy")->kill(deadHarpy), deadHarpy);

    expect.fleet(Side::Defense, 0)->grantSalvage(deadHarpy * stats["Harpy"]->cost * 0.2);
    expect.fleet(Side::Defense, 0)->grantSalvage(deadPegasus * stats["Pegasus"]->cost * 0.4);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void sameTimeKill() {
    // Ranger and Wyvern shoot at each other at the same time for full damage.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Ranger"], 3500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Wyvern"],  200));

    Battle expect(actual);
    NumberOfShips deadRanger =  200 * stats["Wyvern"]->guns / stats["Ranger"]->armor;
    NumberOfShips deadWyvern = 3500 * stats["Ranger"]->guns / stats["Wyvern"]->armor;
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Ranger")->kill(deadRanger), deadRanger);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Wyvern")->kill(deadWyvern), deadWyvern);

    expect.fleet(Side::Defense, 0)->grantSalvage(deadRanger * stats["Ranger"]->cost * 0.2);
    expect.fleet(Side::Defense, 0)->grantSalvage(deadWyvern * stats["Wyvern"]->cost * 0.4);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void sameTimeFullKill() {
    // Avenger and Vindicator shoot at each other at the same time for full
    // damage. Vindicator kill some Avenger with some overkill, Avenger kill all
    // Vindicator with some overkill.
    Battle actual(1, 1);
    NumberOfShips startVindicator = 600 * stats["Avenger"]->armor / stats["Vindicator"]->guns + 1;
    NumberOfShips startAvenger = 1000;
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Avenger"],    startAvenger));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],                 0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Vindicator"], startVindicator));

    Battle expect(actual);
    NumberOfShips deadVindicator =    startAvenger * stats["Avenger"]->guns    / stats["Vindicator"]->armor;
    NumberOfShips deadAvenger    = startVindicator * stats["Vindicator"]->guns / stats["Avenger"]->armor;
    EXPECT_TRUE(deadVindicator > startVindicator);
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Avenger")->kill(deadAvenger), deadAvenger);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Vindicator")->kill(deadVindicator), startVindicator);

    expect.fleet(Side::Defense, 0)->grantSalvage(deadAvenger     * stats["Avenger"]->cost * 0.2);
    expect.fleet(Side::Defense, 0)->grantSalvage(startVindicator * stats["Vindicator"]->cost * 0.4);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void overshootKill() {
    // Phantom do not fire at Pegasus. Pegasus more than kill all Phantom.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Phantom"], 5000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],   0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Pegasus"], 1000));

    Battle expect(actual);
    NumberOfShips startPhantom = expect.fleet(Side::Attack, 0)->battleFormation("Phantom")->atStartOfBattle;
    NumberOfShips deadPhantom = 1000 * stats["Pegasus"]->guns / stats["Phantom"]->armor;
    EXPECT_TRUE(deadPhantom > startPhantom);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Phantom")->kill(deadPhantom), startPhantom);

    expect.fleet(Side::Defense, 0)->grantSalvage(startPhantom * stats["Phantom"]->cost * 0.2);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void overshootRepeatKill() {
    // Phantom do not fire at Pegasus. Pegasus more than kill all Phantom and
    // some Thief. Overkill guns then reallocated to Thief, some more or which
    // are killed.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Phantom"], 5000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Thief"],   5000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],   0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Pegasus"],  600));

    Battle expect(actual);
    NumberOfShips startPhantom = expect.fleet(Side::Attack, 0)->battleFormation("Phantom")->atStartOfBattle;
    NumberOfShips startThief   = expect.fleet(Side::Attack, 0)->battleFormation("Thief")->atStartOfBattle;
    Guns assignedEach = 600 * 5000 / (5000+5000) * stats["Pegasus"]->guns;
    NumberOfShips deadPhantom = assignedEach / stats["Phantom"]->armor;
    NumberOfShips deadThief0   = assignedEach / stats["Thief"]->armor;
    EXPECT_TRUE(deadPhantom > startPhantom);
    EXPECT_TRUE(deadThief0 < startThief);
    Guns neededForPhantom = startPhantom * stats["Phantom"]->armor;
    NumberOfShips deadThief1 = (assignedEach - neededForPhantom) / stats["Thief"]->armor;
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Phantom")->kill(deadPhantom), startPhantom);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Thief")->kill(deadThief0 + deadThief1), deadThief0 + deadThief1);

    expect.fleet(Side::Defense, 0)->grantSalvage((startPhantom * stats["Phantom"]->cost +
                                                  deadThief0 * stats["Thief"]->cost +
                                                  deadThief1 * stats["Thief"]->cost) * 0.2);

    bool useRepeatSameTargetIndex = true;
    bcalc(actual, useRepeatSameTargetIndex);
    EXPECT_EQUAL(actual, expect);
}

void t2Kill() {
    // Phoenix first kill all Spectre, then some Ranger. Ranger and Spectre have
    // no targets.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack, 0, *stats["Ranger"],    500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack, 0, *stats["Spectre"],   300));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],   0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Phoenix"], 1300));

    Battle expect(actual);
    NumberOfShips deadSpectre = actual.fleet(Side::Attack, 0)->battleFormation("Spectre")->atStartOfBattle;
    Guns t2Guns = 1300 * stats["Phoenix"]->guns - 300 * stats["Spectre"]->armor;
    NumberOfShips deadRanger = t2Guns * effs[1] / stats["Ranger"]->armor / 100;
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Spectre")->kill(deadSpectre), deadSpectre);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Ranger" )->kill(deadRanger),  deadRanger);

    expect.fleet(Side::Defense, 0)->grantSalvage((deadSpectre * stats["Spectre"]->cost +
                                                  deadRanger * stats["Ranger"]->cost) * 0.2);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void t2AfterOvershootKill() {
    // Spider and Beetle have no target. Pegasus kill some Spider and overkill
    // all Phantom. Overkill guns are reallocated to kill some Beetle. All
    // Phantom are dead before they can fire and have no target, anyway.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Phantom"], 2000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Spider"],  1000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Beetle"],   500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],   0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Pegasus"],  255));

    Battle expect(actual);
    Guns phantomGuns = 255 * 2000 * stats["Pegasus"]->guns / (2000+1000);
    NumberOfShips deadPhantom  = phantomGuns / stats["Phantom"]->armor;
    NumberOfShips startPhantom = expect.fleet(Side::Attack,  0)->battleFormation("Phantom")->atStartOfBattle;
    EXPECT_TRUE(deadPhantom > startPhantom);
    NumberOfShips deadSpider    = 255 * 1000 * stats["Pegasus"]->guns / (2000+1000) / stats["Spider"]->armor;

    Guns t2Guns = phantomGuns - 2000 * stats["Phantom"]->armor;
    NumberOfShips deadBeetle  = t2Guns * effs[1] / stats["Beetle"]->armor / 100;
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Phantom")->kill(deadPhantom), startPhantom);
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Spider")->kill(deadSpider), deadSpider);
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Beetle")->kill(deadBeetle), deadBeetle);

    expect.fleet(Side::Defense, 0)->grantSalvage((startPhantom * stats["Phantom"]->cost +
                                                  deadSpider * stats["Spider"]->cost +
                                                  deadBeetle * stats["Beetle"]->cost) * 0.2);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void t2AfterRepeatOvershootKill() {
    // Spider and Beetle have no target. Pegasus kill some Spider and overkill
    // all Phantom. Overkill guns allocated to Phantom are reallocated to
    // overkill remaining Spider. Overkill guns are then reallocated to kill
    // some Beetle at T2. All Phantom are dead before they can fire and have no
    // target, anyway.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Phantom"], 3000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Spider"],  1500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Beetle"],   750));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],   0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Pegasus"],  375));

    Battle expect(actual);
    Guns phantomGuns = 375 * 3000 * stats["Pegasus"]->guns / (3000+1500);
    NumberOfShips deadPhantom  = phantomGuns / stats["Phantom"]->armor;
    NumberOfShips startPhantom = expect.fleet(Side::Attack,  0)->battleFormation("Phantom")->atStartOfBattle;
    EXPECT_TRUE(deadPhantom > startPhantom);
    NumberOfShips deadSpider  = 375 * 1500 * stats["Pegasus"]->guns / (3000+1500) / stats["Spider"]->armor;
    Guns overkillGuns = phantomGuns - startPhantom * stats["Phantom"]->armor;

    NumberOfShips deadSpider2 = overkillGuns / stats["Spider"]->armor;
    NumberOfShips startSpider = expect.fleet(Side::Attack,  0)->battleFormation("Spider")->atStartOfBattle;
    EXPECT_TRUE(deadSpider + deadSpider2 > startSpider);
    Guns overkillGuns2 = overkillGuns - (startSpider - deadSpider) * stats["Spider"]->armor;

    NumberOfShips deadBeetle  = overkillGuns2 * effs[1] / stats["Beetle"]->armor / 100;
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Phantom")->kill(deadPhantom), startPhantom);
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Spider"  )->kill(deadSpider + deadSpider2), startSpider);
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Beetle")->kill(deadBeetle), deadBeetle);

    expect.fleet(Side::Defense, 0)->grantSalvage((startPhantom * stats["Phantom"]->cost +
                                                  startSpider * stats["Spider"]->cost +
                                                  deadBeetle * stats["Beetle"]->cost) * 0.2);

    bool useRepeatSameTargetIndex = true;
    bcalc(actual, useRepeatSameTargetIndex);
    EXPECT_EQUAL(actual, expect);
}


// EMP /////////////////////////////////////////////////////////////////////////

void zeroLossEmp() {
    // Beetle EMP some Phantom. Phantom have no target.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Phantom"], 5000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],   0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Beetle"],  1000));

    Battle expect(actual);
    NumberOfShips huggedPhantom = 1000 * stats["Beetle"]->guns * (100-stats["Phantom"]->eres) / 100.0f;
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Phantom")->emp(huggedPhantom), huggedPhantom);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void distributeEmp() {
    // Both Beetle fleets EMP some Harpy and some Phantom. Harpy and Phantom
    // have no target.
    Battle actual(2, 2);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Harpy"],    6000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  1, *stats["Phantom"], 10000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],    0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Beetle"],   2160));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 1, *stats["Beetle"],   2160));

    Battle expect(actual);
    NumberOfShips huggedHarpy   =  6000*4320/(6000+10000) * stats["Beetle"]->guns * (100-stats["Harpy"]->eres  )/100.0f;
    NumberOfShips huggedPhantom = 10000*4320/(6000+10000) * stats["Beetle"]->guns * (100-stats["Phantom"]->eres) / 100.0f;
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Harpy"  )->emp(huggedHarpy),   huggedHarpy);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 1)->battleFormation("Phantom")->emp(huggedPhantom), huggedPhantom);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void exchangeEmp() {
    // Beetle partially EMP Viper. Remaining Viper partially EMP Beetle.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Viper"],  6000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Beetle"], 2000));

    Battle expect(actual);
    NumberOfShips huggedViper  =  2000                * stats["Beetle"]->guns * (100-stats["Viper"]->eres )/100.0f;
    NumberOfShips huggedBeetle = (6000 - huggedViper) * stats["Viper"]->guns  * (100-stats["Beetle"]->eres) / 100.0f;
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Viper" )->emp(huggedViper),   huggedViper);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Beetle")->emp(huggedBeetle), huggedBeetle);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void sameTimeEmp() {
    // Spider fleets partially EMP each other at the same time.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Spider"], 2000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Spider"], 3000));

    Battle expect(actual);
    NumberOfShips huggedDefSpider = 2000 * stats["Spider"]->guns * (100-stats["Spider"]->eres) / 100.0f;
    NumberOfShips huggedAtkSpider = 3000 * stats["Spider"]->guns * (100-stats["Spider"]->eres) / 100.0f;
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Spider")->emp(huggedAtkSpider), huggedAtkSpider);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Spider")->emp(huggedDefSpider), huggedDefSpider);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void overshootEmp() {
    // Harpy do not fire at Beetle. Beetle more than EMP all Harpy.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Harpy"], 4500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"], 0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Beetle"], 3000));

    Battle expect(actual);
    NumberOfShips startHarpy = expect.fleet(Side::Attack, 0)->battleFormation("Harpy")->atStartOfBattle;
    NumberOfShips huggedHarpy = 3000 * stats["Beetle"]->guns * (100-stats["Harpy"]->eres) / 100.0f;
    EXPECT_TRUE(huggedHarpy > startHarpy);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Harpy")->emp(huggedHarpy), startHarpy);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void overshootRepeatEmp() {
    // Beetle hug some Harpy and overhug all spiders. Overkill guns are
    // reallocated to Harpy, some more of which are hugged.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Harpy"],  4500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Spider"], 3000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Beetle"], 3200));

    Battle expect(actual);
    NumberOfShips huggedHarpy = 3200 * 4500 * stats["Beetle"]->guns * (100-stats["Harpy"]->eres) / 100.0f / (4500+3000);
    Guns spiderGuns = 3200 * 3000 * stats["Beetle"]->guns / (4500+3000);
    NumberOfShips startSpider = expect.fleet(Side::Attack, 0)->battleFormation("Spider")->atStartOfBattle;
    NumberOfShips huggedSpider = spiderGuns * (100-stats["Spider"]->eres) / 100.0f;
    EXPECT_TRUE(huggedSpider > startSpider);
    Guns overkillGuns = spiderGuns - startSpider / (100-stats["Spider"]->eres) * 100.0f;
    NumberOfShips huggedHarpy2 = overkillGuns * (100-stats["Harpy"]->eres) / 100.0f;
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Harpy")->emp(huggedHarpy + huggedHarpy2), huggedHarpy + huggedHarpy2);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Spider")->emp(huggedSpider), startSpider);

    bool useRepeatSameTargetIndex = true;
    bcalc(actual, useRepeatSameTargetIndex);
    EXPECT_EQUAL(actual, expect);
}

void t2Emp() {
    // Viper first EMP all Phoenix, then some Ranger. Phoenix and Ranger have no
    // targets.
    Battle actual(2, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack, 0, *stats["Phoenix"], 600));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack, 1, *stats["Ranger"],  400));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"], 0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Viper"], 1600));

    Battle expect(actual);
    NumberOfShips startPhoenix = actual.fleet(Side::Attack, 0)->battleFormation("Phoenix")->atStartOfBattle;
    NumberOfShips huggedPhoenix = 1600 * stats["Viper"]->guns * std::nextafter(1.0f-stats["Phoenix"]->eres/100.0f, 1.0f);
    EXPECT_TRUE(huggedPhoenix > startPhoenix);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Phoenix")->emp(huggedPhoenix), startPhoenix);

    Guns t2Guns = 1600 * stats["Viper"]->guns - startPhoenix / std::nextafter(1.0f-stats["Phoenix"]->eres/100.0f, 1.0f);
    NumberOfShips huggedRanger = t2Guns * effs[1] * std::nextafter(1.0f-stats["Ranger"]->eres/100.0f, 1.0f) / 100;
    EXPECT_EQUAL(expect.fleet(Side::Attack, 1)->battleFormation("Ranger" )->emp(huggedRanger),  huggedRanger);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void t2AfterOvershootEmp() {
    // Viper first EMP all Cutlass and some Phoenix. Viper then EMP some Ranger.
    // Cutlass is fully EMPed, Phoenix and Ranger have no targets.
    Battle actual(3, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack, 0, *stats["Cutlass"], 1200));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack, 1, *stats["Phoenix"],  600));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack, 2, *stats["Ranger"],   400));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Viper"],  3300));

    Battle expect(actual);
    NumberOfShips startCutlass = actual.fleet(Side::Attack, 0)->battleFormation("Cutlass")->atStartOfBattle;
    NumberOfShips huggedCutlass = 3300 * 1200 / (1200+600) * stats["Viper"]->guns * std::nextafter(1.0f-stats["Cutlass"]->eres/100.0f, 1.0f);
    EXPECT_TRUE(huggedCutlass > startCutlass);
    NumberOfShips huggedPhoenix = 3300 *  600 / (1200+600) * stats["Viper"]->guns * std::nextafter(1.0f-stats["Phoenix"]->eres/100.0f, 1.0f);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Cutlass")->emp(huggedCutlass), startCutlass);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 1)->battleFormation("Phoenix")->emp(huggedPhoenix), huggedPhoenix);

    Guns t2Guns = 3300 * 1200 / (1200+600) * stats["Viper"]->guns - (1200 / std::nextafter(1.0f-stats["Cutlass"]->eres/100.0f, 1.0f));
    NumberOfShips huggedRanger = t2Guns * effs[1] * std::nextafter(1.0f-stats["Ranger"]->eres/100.0f, 1.0f)/ 100;
    EXPECT_EQUAL(expect.fleet(Side::Attack, 2)->battleFormation("Ranger" )->emp(huggedRanger),  huggedRanger);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void t2AfterRepeatOvershootEmp() {
    // Viper first EMP all Cutlass and some Phoenix. Viper then EMP the
    // remaining Phoenix, and finally some Ranger. Cutlass is fully EMPed,
    // Phoenix and Ranger have no targets.
    Battle actual(3, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack, 0, *stats["Cutlass"], 1200));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack, 1, *stats["Phoenix"],  600));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack, 2, *stats["Ranger"],   400));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Viper"],  3300));

    Battle expect(actual);
    NumberOfShips startPhoenix = actual.fleet(Side::Attack, 1)->battleFormation("Phoenix")->atStartOfBattle;
    NumberOfShips huggedPhoenix = 3300 *  600 / (1200+600) * stats["Viper"]->guns * std::nextafter(1.0f-stats["Phoenix"]->eres/100.0f, 1.0f);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 1)->battleFormation("Phoenix")->emp(huggedPhoenix), huggedPhoenix);

    NumberOfShips startCutlass = actual.fleet(Side::Attack, 0)->battleFormation("Cutlass")->atStartOfBattle;
    Guns cutlassGuns = 3300 * 1200 / (1200+600) * stats["Viper"]->guns;
    NumberOfShips huggedCutlass = cutlassGuns * std::nextafter(1.0f-stats["Cutlass"]->eres/100.0f, 1.0f);
    EXPECT_TRUE(huggedCutlass > startCutlass);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Cutlass")->emp(huggedCutlass), startCutlass);
    Guns overkillGuns = cutlassGuns - startCutlass / std::nextafter(1.0f-stats["Cutlass"]->eres/100.0f, 1.0f);

    NumberOfShips huggedPhoenix2 = overkillGuns * std::nextafter(1.0f-stats["Phoenix"]->eres/100.0f, 1.0f);
    EXPECT_TRUE(huggedPhoenix2 > startPhoenix - huggedPhoenix);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 1)->battleFormation("Phoenix")->emp(huggedPhoenix2), startPhoenix - huggedPhoenix);
    Guns overkillGuns2 = overkillGuns - (startPhoenix - huggedPhoenix) / std::nextafter(1.0f-stats["Phoenix"]->eres/100.0f, 1.0f);

    NumberOfShips huggedRanger = overkillGuns2 * effs[1] * std::nextafter(1.0f-stats["Ranger"]->eres/100.0f, 1.0f) / 100;
    EXPECT_EQUAL(expect.fleet(Side::Attack, 2)->battleFormation("Ranger" )->emp(huggedRanger),  huggedRanger);

    bool useRepeatSameTargetIndex = true;
    bcalc(actual, useRepeatSameTargetIndex);
    EXPECT_EQUAL(actual, expect);
}


// STEAL ///////////////////////////////////////////////////////////////////////

void zeroLossSteal() {
    // Harpy have no target. Cutlass steal some Harpy.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Harpy"],  2000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Cutlass"], 500));

    Battle expect(actual);
    NumberOfShips stolenHarpy = 500 * stats["Cutlass"]->guns / stats["Harpy"]->armor;
    NumberOfShips abandonedCutlass = stolenHarpy * stats["Harpy"]->cost / stats["Cutlass"]->cost;
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Harpy")->steal(stolenHarpy), stolenHarpy);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Cutlass")->abandon(abandonedCutlass), abandonedCutlass);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Harpy"], stolenHarpy);
    expect.fleet(Side::Defense, 0)->grantSalvage(stolenHarpy * stats["Harpy"]->cost * 0.2f);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void stolenShipsUntargetableSteal() {
    // Cutlass steal some Harpy. Stolen Harpy cannot be fired upon by Buccaneer.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Harpy"],     2000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Buccaneer"], 1000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],     0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Cutlass"],    500));

    Battle expect(actual);
    NumberOfShips stolenHarpy = 500 * stats["Cutlass"]->guns / stats["Harpy"]->armor;
    NumberOfShips abandonedCutlass = stolenHarpy * stats["Harpy"]->cost / stats["Cutlass"]->cost;
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Harpy")->steal(stolenHarpy), stolenHarpy);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Cutlass")->abandon(abandonedCutlass), abandonedCutlass);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Harpy"], stolenHarpy);

    expect.fleet(Side::Defense, 0)->grantSalvage(stolenHarpy * stats["Harpy"]->cost * 0.2);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void distributeSteal() {
    // Harpy and Phantom have no target. Cutlass fleets steal some Harpy and
    // some Phantom.
    Battle actual(1, 2);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Harpy"],   3000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Phantom"], 3000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],   0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Cutlass"],  800));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 1, *stats["Cutlass"],  400));

    Battle expect(actual);
    NumberOfShips stolenHarpy       = (800 + 400) * stats["Cutlass"]->guns / stats["Harpy"]->armor   / 2;
    NumberOfShips stolenPhantom     = (800 + 400) * stats["Cutlass"]->guns / stats["Phantom"]->armor / 2;
    NumberOfShips abandonedCutlass0 = stolenHarpy * 2 / 3 * stats["Harpy"]->cost / stats["Cutlass"]->cost +
        stolenPhantom * 2 / 3 * stats["Phantom"]->cost / stats["Cutlass"]->cost;
    NumberOfShips abandonedCutlass1 = stolenHarpy * 1 / 3 * stats["Harpy"]->cost / stats["Cutlass"]->cost +
        stolenPhantom * 1 / 3 * stats["Phantom"]->cost / stats["Cutlass"]->cost;
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Harpy"  )->steal(stolenHarpy),         stolenHarpy);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Cutlass")->abandon(abandonedCutlass0), abandonedCutlass0);
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Phantom")->steal(stolenPhantom),       stolenPhantom);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 1)->battleFormation("Cutlass")->abandon(abandonedCutlass1), abandonedCutlass1);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Harpy"],   stolenHarpy   * 2 / 3);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Phantom"], stolenPhantom * 2 / 3);
    expect.fleet(Side::Defense, 1)->addStolen(*stats["Harpy"],   stolenHarpy       / 3);
    expect.fleet(Side::Defense, 1)->addStolen(*stats["Phantom"], stolenPhantom     / 3);

    expect.fleet(Side::Defense, 0)->grantSalvage(stolenHarpy   * stats["Harpy"]->cost   * 2 / 3 * 0.2 +
                                                 stolenPhantom * stats["Phantom"]->cost * 2 / 3 * 0.2);
    expect.fleet(Side::Defense, 1)->grantSalvage(stolenHarpy   * stats["Harpy"]->cost       / 3 * 0.2 +
                                                 stolenPhantom * stats["Phantom"]->cost     / 3 * 0.2);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void exchangeSteal() {
    // Cutlass steal some Thief. Thief steal some Cutlass.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Thief"],   2500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],   0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Cutlass"], 1000));

    Battle expect(actual);
    NumberOfShips stolenThief   = 1000                 * stats["Cutlass"]->guns / stats["Thief"]->armor;
    NumberOfShips stolenCutlass = (2500 - stolenThief) * stats["Thief"]->guns   / stats["Cutlass"]->armor;
    NumberOfShips abandonedCutlass = stolenThief   * stats["Thief"]->cost / stats["Cutlass"]->cost;
    NumberOfShips abandonedThief   = stolenCutlass * stats["Cutlass"]->cost / stats["Thief"]->cost;
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Thief"  )->steal(  stolenThief),      stolenThief);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Cutlass")->abandon(abandonedCutlass), abandonedCutlass);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Cutlass")->steal(  stolenCutlass),    stolenCutlass);
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Thief"  )->abandon(abandonedThief),   abandonedThief);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Thief"],   stolenThief);
    expect.fleet(Side::Attack,  0)->addStolen(*stats["Cutlass"], stolenCutlass);

    expect.fleet(Side::Defense, 0)->grantSalvage(stolenThief   * stats["Thief"]->cost * 0.2);
    expect.fleet(Side::Defense, 0)->grantSalvage(stolenCutlass * stats["Cutlass"]->cost * 0.4);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void sameTimeSteal() {
    // Thief steal some Cutter and Cutter steal some Thief, at the same time.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Thief"],  2500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Cutter"], 1000));

    Battle expect(actual);
    NumberOfShips startThief  = actual.fleet(Side::Attack,  0)->battleFormation("Thief")->atStartOfBattle;
    NumberOfShips startCutter = actual.fleet(Side::Defense, 0)->battleFormation("Cutter")->atStartOfBattle;
    NumberOfShips stolenThief  = 1000 * stats["Cutter"]->guns / stats["Thief"]->armor;
    NumberOfShips stolenCutter = 2500 * stats["Thief"]->guns  / stats["Cutter"]->armor;
    NumberOfShips abandonedCutter = stolenThief   * stats["Thief"]->cost / stats["Cutter"]->cost;
    NumberOfShips abandonedThief  = stolenCutter * stats["Cutter"]->cost / stats["Thief"]->cost;
    EXPECT_TRUE(stolenThief + abandonedThief <= startThief);
    EXPECT_TRUE(stolenCutter + abandonedCutter <= startCutter);
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Thief" )->steal(  stolenThief),     stolenThief);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Cutter")->abandon(abandonedCutter), abandonedCutter);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Cutter")->steal(  stolenCutter),    stolenCutter);
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Thief" )->abandon(abandonedThief),  abandonedThief);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Thief"],  stolenThief);
    expect.fleet(Side::Attack,  0)->addStolen(*stats["Cutter"], stolenCutter);

    expect.fleet(Side::Defense, 0)->grantSalvage(stolenCutter * stats["Cutter"]->cost * 0.4);
    expect.fleet(Side::Defense, 0)->grantSalvage(stolenThief * stats["Thief"]->cost * 0.2);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void overshootSteal() {
    // Cutlass steal some Harpy and all Phantom. Cutlass then continue to steal
    // more Harpy.
    Battle actual(2, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Harpy"],   2000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  1, *stats["Phantom"], 4000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],   0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Cutlass"], 3600));

    Battle expect(actual);
    NumberOfShips startPhantom = actual.fleet(Side::Attack, 1)->battleFormation("Phantom")->atStartOfBattle;
    NumberOfShips stolenHarpy   = 2000 * 3600 / (2000+4000) * stats["Cutlass"]->guns / stats["Harpy"]->armor;
    NumberOfShips stolenPhantom = 4000 * 3600 / (2000+4000) * stats["Cutlass"]->guns / stats["Phantom"]->armor;
    EXPECT_TRUE(stolenPhantom > startPhantom);
    NumberOfShips abandonedCutlass = stolenHarpy  * stats["Harpy"]->cost   / stats["Cutlass"]->cost +
        startPhantom * stats["Phantom"]->cost / stats["Cutlass"]->cost;
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Harpy"  )->steal(  stolenHarpy), stolenHarpy);
    EXPECT_EQUAL(expect.fleet(Side::Attack,  1)->battleFormation("Phantom")->steal(stolenPhantom), startPhantom);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Cutlass")->abandon(abandonedCutlass), abandonedCutlass);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Harpy"],   stolenHarpy);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Phantom"], startPhantom);

    expect.fleet(Side::Defense, 0)->grantSalvage(startPhantom * stats["Phantom"]->cost * 0.2 +
                                                 stolenHarpy  * stats["Harpy"]->cost * 0.2);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void overshootRepeatSteal() {
    // Cutlass kill some Harpy and all Phantom. Cutlass then continue to steal
    // more Harpy.
    Battle actual(2, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Harpy"],   2000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  1, *stats["Phantom"], 4000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],   0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Cutlass"], 3600));

    Battle expect(actual);
    NumberOfShips startPhantom = actual.fleet(Side::Attack, 1)->battleFormation("Phantom")->atStartOfBattle;
    NumberOfShips stolenHarpy   = 2000 * 3600 / (2000+4000) * stats["Cutlass"]->guns / stats["Harpy"]->armor;
    Guns phantomGuns = 4000 * 3600 / (2000+4000) * stats["Cutlass"]->guns;
    NumberOfShips stolenPhantom = phantomGuns / stats["Phantom"]->armor;
    EXPECT_TRUE(stolenPhantom > startPhantom);
    Guns overkillGuns = phantomGuns - startPhantom * stats["Phantom"]->armor;
    NumberOfShips stolenHarpy2 = overkillGuns / stats["Harpy"]->armor;
    NumberOfShips abandonedCutlass = stolenHarpy  * stats["Harpy"]->cost   / stats["Cutlass"]->cost +
        stolenHarpy2 * stats["Harpy"]->cost   / stats["Cutlass"]->cost +
        startPhantom * stats["Phantom"]->cost / stats["Cutlass"]->cost;
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Harpy"  )->steal(stolenHarpy + stolenHarpy2), stolenHarpy + stolenHarpy2);
    EXPECT_EQUAL(expect.fleet(Side::Attack,  1)->battleFormation("Phantom")->steal(stolenPhantom), startPhantom);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Cutlass")->abandon(abandonedCutlass), abandonedCutlass);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Harpy"],   stolenHarpy + stolenHarpy2);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Phantom"], startPhantom);

    expect.fleet(Side::Defense, 0)->grantSalvage(startPhantom * stats["Phantom"]->cost * 0.2 +
                                                 (stolenHarpy + stolenHarpy2) * stats["Harpy"]->cost * 0.2);

    bool useRepeatSameTargetIndex = true;
    bcalc(actual, useRepeatSameTargetIndex);
    EXPECT_EQUAL(actual, expect);
}

void t2Steal() {
    // Phantom and Phoenix have no targets. Cutlass steal all Phantom at T1 and
    // some Phoenix at T2.
    Battle actual(2, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Phantom"], 4000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  1, *stats["Phoenix"], 1000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],   0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Cutlass"], 3600));

    Battle expect(actual);
    NumberOfShips startPhantom = actual.fleet(Side::Attack, 0)->battleFormation("Phantom")->atStartOfBattle;
    NumberOfShips stolenPhantom = 3600 * stats["Cutlass"]->guns / stats["Phantom"]->armor;
    EXPECT_TRUE(stolenPhantom > startPhantom);
    Guns t2Guns = 3600 * stats["Cutlass"]->guns - startPhantom * stats["Phantom"]->armor;
    NumberOfShips stolenPhoenix = t2Guns * effs[1] / stats["Phoenix"]->armor / 100;
    NumberOfShips abandonedCutlass = startPhantom  * stats["Phantom"]->cost / stats["Cutlass"]->cost +
        stolenPhoenix * stats["Phoenix"]->cost / stats["Cutlass"]->cost;
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Phantom")->steal(stolenPhantom), startPhantom);
    EXPECT_EQUAL(expect.fleet(Side::Attack,  1)->battleFormation("Phoenix")->steal(stolenPhoenix), stolenPhoenix);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Cutlass")->abandon(abandonedCutlass), abandonedCutlass);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Phantom"], startPhantom);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Phoenix"], stolenPhoenix);

    expect.fleet(Side::Defense, 0)->grantSalvage(startPhantom * stats["Phantom"]->cost * 0.2 +
                                                 stolenPhoenix * stats["Phoenix"]->cost * 0.2 );

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void t2AfterRepeatOvershootSteal() {
    // Harpy, Phantom and Phoenix have no targets. Cutlass steal some Harpy and
    // all Phantom at T1. Guns that overkill Phantom are reassigned to Harpy,
    // some more of which are stolen, still at T1.
    Battle actual(3, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Harpy"],   2000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  1, *stats["Phantom"], 4000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  2, *stats["Phoenix"], 1000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],   0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Cutlass"], 3600));

    Battle expect(actual);
    NumberOfShips stolenHarpy   = 2000 * 3600 / (2000+4000) * stats["Cutlass"]->guns / stats["Harpy"]->armor;
    NumberOfShips startPhantom = actual.fleet(Side::Attack, 1)->battleFormation("Phantom")->atStartOfBattle;
    Guns phantomGuns = 4000 * 3600 / (2000+4000) * stats["Cutlass"]->guns;
    NumberOfShips stolenPhantom = phantomGuns / stats["Phantom"]->armor;
    EXPECT_TRUE(stolenPhantom > startPhantom);
    Guns overkillGuns = phantomGuns - startPhantom * stats["Phantom"]->armor;
    NumberOfShips stolenHarpy2 = overkillGuns / stats["Harpy"]->armor;
    NumberOfShips abandonedCutlass = stolenHarpy  * stats["Harpy"]->cost   / stats["Cutlass"]->cost +
        stolenHarpy2 * stats["Harpy"]->cost   / stats["Cutlass"]->cost +
        startPhantom * stats["Phantom"]->cost / stats["Cutlass"]->cost;
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Harpy"  )->steal(stolenHarpy + stolenHarpy2), stolenHarpy + stolenHarpy2);
    EXPECT_EQUAL(expect.fleet(Side::Attack,  1)->battleFormation("Phantom")->steal(stolenPhantom), startPhantom);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Cutlass")->abandon(abandonedCutlass), abandonedCutlass);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Harpy"],   stolenHarpy + stolenHarpy2);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Phantom"], startPhantom);

    expect.fleet(Side::Defense, 0)->grantSalvage((stolenHarpy + stolenHarpy2) * stats["Harpy"]->cost   * 0.2 +
                                                 startPhantom                 * stats["Phantom"]->cost * 0.2);

    bool useRepeatSameTargetIndex = true;
    bcalc(actual, useRepeatSameTargetIndex);
    EXPECT_EQUAL(actual, expect);
}

void t2AfterOvershootSteal() {
    // Harpy, Phantom and Phoenix have no targets. Cutlass steal some Harpy and
    // all Phantom at T1. Guns that overkill Phantom are reassigned to Phoenix,
    // some of which are then stolen at T2.
    Battle actual(3, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Harpy"],   2000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  1, *stats["Phantom"], 4000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  2, *stats["Phoenix"], 1000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],   0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Cutlass"], 3600));

    Battle expect(actual);
    NumberOfShips startPhantom = actual.fleet(Side::Attack, 1)->battleFormation("Phantom")->atStartOfBattle;
    NumberOfShips stolenHarpy   = 2000 * 3600 / (2000+4000) * stats["Cutlass"]->guns / stats["Harpy"]->armor;
    Guns t1Guns = 4000 * 3600 / (2000+4000) * stats["Cutlass"]->guns;
    NumberOfShips stolenPhantom = t1Guns / stats["Phantom"]->armor;
    EXPECT_TRUE(stolenPhantom > startPhantom);
    Guns t2Guns = t1Guns - startPhantom * stats["Phantom"]->armor;
    NumberOfShips stolenPhoenix = t2Guns * effs[1] / stats["Phoenix"]->armor / 100;
    NumberOfShips abandonedCutlass = stolenHarpy   * stats["Harpy"]->cost   / stats["Cutlass"]->cost +
        startPhantom  * stats["Phantom"]->cost / stats["Cutlass"]->cost +
        stolenPhoenix * stats["Phoenix"]->cost / stats["Cutlass"]->cost;
    EXPECT_EQUAL(expect.fleet(Side::Attack,  0)->battleFormation("Harpy"  )->steal(  stolenHarpy), stolenHarpy);
    EXPECT_EQUAL(expect.fleet(Side::Attack,  1)->battleFormation("Phantom")->steal(stolenPhantom), startPhantom);
    EXPECT_EQUAL(expect.fleet(Side::Attack,  2)->battleFormation("Phoenix")->steal(stolenPhoenix), stolenPhoenix);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Cutlass")->abandon(abandonedCutlass), abandonedCutlass);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Harpy"],   stolenHarpy);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Phantom"], startPhantom);
    expect.fleet(Side::Defense, 0)->addStolen(*stats["Phoenix"], stolenPhoenix);

    expect.fleet(Side::Defense, 0)->grantSalvage(startPhantom  * stats["Phantom"]->cost * 0.2 +
                                                 stolenHarpy   * stats["Harpy"]->cost   * 0.2 +
                                                 stolenPhoenix * stats["Phoenix"]->cost * 0.2);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}


// POD /////////////////////////////////////////////////////////////////////////

void maxCap() {
    // Enough Illusion to maxcap on target.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Illusion"], 1500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  500));
    actual.fleet(Side::Attack,  0)->value = 400000;
    actual.fleet(Side::Attack,  0)->score = 600000;
    actual.fleet(Side::Defense, 0)->value = 500000;
    actual.fleet(Side::Defense, 0)->score = 800000;

    Battle expect(actual);
    NumberOfShips canCap = 1500 * stats["Illusion"]->guns / stats["Asteroid"]->armor;
    NumberOfShips mayCap = 500 * 0.25 * std::min(1.0f, std::sqrt(500000.0f / 400000));
    EXPECT_TRUE(canCap >= mayCap);
    expect.fleet(Side::Attack,  0)->addStolen(*stats["Asteroid"], mayCap);
    float bravery = std::max(0.2f, std::min(2.2f, float(expect.fleet(Side::Defense, 0)->score) / expect.fleet(Side::Attack, 0)->score) - 0.2f) *
        std::max(0.2f, std::min(1.8f, float(expect.fleet(Side::Defense, 0)->value) / expect.fleet(Side::Attack,  0)->value) - 0.1f) /
        std::max(1.0f, 0.6f + float(expect.fleet(Side::Attack,  0)->score) / expect.fleet(Side::Attack,  0)->value / 10);
    expect.fleet(Side::Attack,  0)->addXP(mayCap * 10 * bravery);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Asteroid")->cap(canCap), mayCap);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void notEnoughPodsCap() {
    // Not enough Illusion to maxcap on target.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Illusion"], 1000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  500));
    actual.fleet(Side::Attack,  0)->value = 300000;
    actual.fleet(Side::Attack,  0)->score = 500000;
    actual.fleet(Side::Defense, 0)->value = 400000;
    actual.fleet(Side::Defense, 0)->score = 700000;

    Battle expect(actual);
    NumberOfShips mayCap = 500 * 0.25 * std::min(1.0f, std::sqrt(400000.0f / 300000));
    NumberOfShips canCap = 1000 * stats["Illusion"]->guns / stats["Asteroid"]->armor;
    EXPECT_TRUE(canCap < mayCap);
    expect.fleet(Side::Attack,  0)->addStolen(*stats["Asteroid"], canCap);
    float bravery = std::max(0.2f, std::min(2.2f, float(expect.fleet(Side::Defense, 0)->score) / expect.fleet(Side::Attack, 0)->score) - 0.2f) *
        std::max(0.2f, std::min(1.8f, float(expect.fleet(Side::Defense, 0)->value) / expect.fleet(Side::Attack,  0)->value) - 0.1f) /
        std::max(1.0f, 0.6f + float(expect.fleet(Side::Attack,  0)->score) / expect.fleet(Side::Attack,  0)->value / 10);
    expect.fleet(Side::Attack,  0)->addXP(canCap * 10 * bravery);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Asteroid")->cap(canCap), canCap);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void deadPodsCantCap() {
    // Pegasus kill some Illusion, leaving not enough to maxcap on target.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Illusion"], 1250));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Pegasus"],    50));
    actual.fleet(Side::Attack,  0)->value = 100000;
    actual.fleet(Side::Attack,  0)->score = 150000;
    actual.fleet(Side::Defense, 0)->value = 150000;
    actual.fleet(Side::Defense, 0)->score = 200000;

    Battle expect(actual);
    NumberOfShips deadIllusion = 50 * stats["Pegasus"]->guns / stats["Illusion"]->armor;
    EXPECT_TRUE(deadIllusion < 1250);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Illusion")->kill(deadIllusion), deadIllusion);

    NumberOfShips mayCap = 500 * 0.25 * std::min(1.0f, std::sqrt(150000.0f / 100000));
    NumberOfShips canCap = (1250 - deadIllusion) * stats["Illusion"]->guns / stats["Asteroid"]->armor;
    EXPECT_TRUE(canCap < mayCap);
    expect.fleet(Side::Attack,  0)->addStolen(*stats["Asteroid"], canCap);
    float bravery = std::max(0.2f, std::min(2.2f, float(expect.fleet(Side::Defense, 0)->score) / expect.fleet(Side::Attack, 0)->score) - 0.2f) *
        std::max(0.2f, std::min(1.8f, float(expect.fleet(Side::Defense, 0)->value) / expect.fleet(Side::Attack,  0)->value) - 0.1f) /
        std::max(1.0f, 0.6f + float(expect.fleet(Side::Attack,  0)->score) / expect.fleet(Side::Attack,  0)->value / 10);
    expect.fleet(Side::Attack,  0)->addXP(canCap * 10 * bravery);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Asteroid")->cap(canCap), canCap);

    Cost salvage = deadIllusion * stats["Illusion"]->cost * 0.2f;
    Cost salvageCap = expect.fleet(Side::Defense, 0)->fleetCost() / 5;
    EXPECT_TRUE(salvage > salvageCap);
    expect.fleet(Side::Defense, 0)->grantSalvage(salvageCap);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void stillAlivePodsCap() {
    // Illusion cap roids first, then die.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Illusion"], 1250));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Clipper"],    50));
    actual.fleet(Side::Attack,  0)->value = 100000;
    actual.fleet(Side::Attack,  0)->score = 150000;
    actual.fleet(Side::Defense, 0)->value = 150000;
    actual.fleet(Side::Defense, 0)->score = 200000;

    Battle expect(actual);
    NumberOfShips mayCap =  500 * 0.25 * std::min(1.0f, std::sqrt(150000.0f / 100000));
    NumberOfShips canCap = 1250 * stats["Illusion"]->guns / stats["Asteroid"]->armor;
    EXPECT_TRUE(canCap == mayCap);
    expect.fleet(Side::Attack,  0)->addStolen(*stats["Asteroid"], canCap);
    float bravery = std::max(0.2f, std::min(2.2f, float(expect.fleet(Side::Defense, 0)->score) / expect.fleet(Side::Attack, 0)->score) - 0.2f) *
        std::max(0.2f, std::min(1.8f, float(expect.fleet(Side::Defense, 0)->value) / expect.fleet(Side::Attack,  0)->value) - 0.1f) /
        std::max(1.0f, 0.6f + float(expect.fleet(Side::Attack,  0)->score) / expect.fleet(Side::Attack,  0)->value / 10);
    expect.fleet(Side::Attack,  0)->addXP(canCap * 10 * bravery);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Asteroid")->cap(canCap), canCap);

    NumberOfShips deadIllusion = 50 * stats["Clipper"]->guns / stats["Illusion"]->armor;
    EXPECT_TRUE(deadIllusion < 1250);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Illusion")->kill(deadIllusion), deadIllusion);

    Cost salvage = deadIllusion * stats["Illusion"]->cost * 0.2;
    Cost salvageCap = expect.fleet(Side::Defense, 0)->fleetCost() / 5;
    EXPECT_TRUE(salvage > salvageCap);
    expect.fleet(Side::Defense, 0)->grantSalvage(salvageCap);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void emptyFleetsCantCap() {
    // Illusion has 3X value and Clipper fleet has 2X value. Illusion fleet is
    // therefore entitled to 15% cap and Clipper to 10% cap. Clipper fleet
    // doesn't claim its part of the pie, so it is lost.
    Battle actual(2, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Illusion"], 1500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  1, *stats["Clipper"],   100));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  500));
    actual.fleet(Side::Attack,  0)->value = 100000;
    actual.fleet(Side::Attack,  0)->score = 150000;
    actual.fleet(Side::Attack,  1)->value = 100000;
    actual.fleet(Side::Attack,  1)->score = 150000;
    actual.fleet(Side::Defense, 0)->value = 150000;
    actual.fleet(Side::Defense, 0)->score = 200000;

    Battle expect(actual);
    // Cap limited by amount of value in Illusion fleet.
    NumberOfShips mayCap =  500 * std::min(0.15, 0.15 * std::sqrt(150000.0f / 100000));
    NumberOfShips canCap = 1500 * stats["Illusion"]->guns / stats["Asteroid"]->armor;
    EXPECT_TRUE(canCap > mayCap);
    expect.fleet(Side::Attack,  0)->addStolen(*stats["Asteroid"], mayCap);
    float bravery = std::max(0.2f, std::min(2.2f, float(expect.fleet(Side::Defense, 0)->score) / expect.fleet(Side::Attack, 0)->score) - 0.2f) *
        std::max(0.2f, std::min(1.8f, float(expect.fleet(Side::Defense, 0)->value) / expect.fleet(Side::Attack,  0)->value) - 0.1f) /
        std::max(1.0f, 0.6f + float(expect.fleet(Side::Attack,  0)->score) / expect.fleet(Side::Attack,  0)->value / 10);
    expect.fleet(Side::Attack,  0)->addXP(mayCap * 10 * bravery);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Asteroid")->cap(mayCap), mayCap);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void multipleTypesCap() {
    // What if my fleet contains multiple pod types?
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Illusion"], 500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Ziz"],       50));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Merchant"],  60));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"], 500));
    actual.fleet(Side::Attack,  0)->value = 100000;
    actual.fleet(Side::Attack,  0)->score = 150000;
    actual.fleet(Side::Defense, 0)->value = 150000;
    actual.fleet(Side::Defense, 0)->score = 200000;

    Battle expect(actual);
    // Cap limited by amount of value in Illusion fleet.
    NumberOfShips mayCap =  500 * 0.25 * std::min(1.0f, std::sqrt(150000.0f / 100000));
    NumberOfShips canCap = (500 * stats["Illusion"]->guns + 50 * stats["Ziz"]->guns + 60 * stats["Merchant"]->guns) / stats["Asteroid"]->armor;
    EXPECT_TRUE(canCap > mayCap);
    expect.fleet(Side::Attack,  0)->addStolen(*stats["Asteroid"], mayCap);
    float bravery = std::max(0.2f, std::min(2.2f,  float(expect.fleet(Side::Defense, 0)->score) / expect.fleet(Side::Attack, 0)->score) - 0.2f) *
        std::max(0.2f, std::min(1.8f,  float(expect.fleet(Side::Defense, 0)->value) / expect.fleet(Side::Attack, 0)->value) - 0.1f) /
        std::max(1.0f,          0.6f + float(expect.fleet(Side::Attack,  0)->score) / expect.fleet(Side::Attack, 0)->value / 10);
    expect.fleet(Side::Attack,  0)->addXP(mayCap * 10 * bravery);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Asteroid")->cap(mayCap), mayCap);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void variableLimitCap() {
    // Illusion can maxcap, but the cap is reduced because the attacker's planet
    // value is higher than defender's planet value.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Illusion"], 1500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  500));
    actual.fleet(Side::Attack,  0)->value = 400000;
    actual.fleet(Side::Attack,  0)->score = 700000;
    actual.fleet(Side::Defense, 0)->value = 300000;
    actual.fleet(Side::Defense, 0)->score = 500000;

    Battle expect(actual);
    // Cap limited by amount of value in Illusion fleet.
    float capRate = 0.25 * std::min(1.0f, std::sqrt(300000.0f / 400000));
    EXPECT_TRUE(capRate < 0.25);
    NumberOfShips mayCap =  500 * capRate;
    NumberOfShips canCap = 1500 * stats["Illusion"]->guns / stats["Asteroid"]->armor;
    EXPECT_TRUE(canCap > mayCap);
    expect.fleet(Side::Attack,  0)->addStolen(*stats["Asteroid"], mayCap);
    float bravery = std::max(0.2f, std::min(2.2f, float(expect.fleet(Side::Defense, 0)->score) / expect.fleet(Side::Attack, 0)->score) - 0.2f) *
        std::max(0.2f, std::min(1.8f, float(expect.fleet(Side::Defense, 0)->value) / expect.fleet(Side::Attack,  0)->value) - 0.1f) /
        std::max(1.0f, 0.6f + float(expect.fleet(Side::Attack,  0)->score) / expect.fleet(Side::Attack,  0)->value / 10);
    expect.fleet(Side::Attack,  0)->addXP(mayCap * 10 * bravery);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Asteroid")->cap(mayCap), mayCap);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void distributeCap() {
    // 2 fleets with 2 pods types each maxcap on target. Roids divided among
    // attacking fleets according to their total fleet values.
    Battle actual(2, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Illusion"], 700));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Ziz"],       30));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  1, *stats["Illusion"], 400));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  1, *stats["Merchant"],  80));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"], 500));
    actual.fleet(Side::Attack,  0)->value = 200000;
    actual.fleet(Side::Attack,  0)->score = 400000;
    actual.fleet(Side::Attack,  1)->value = 200000;
    actual.fleet(Side::Attack,  1)->score = 400000;
    actual.fleet(Side::Defense, 0)->value = 300000;
    actual.fleet(Side::Defense, 0)->score = 500000;

    Battle expect(actual);
    // Cap limited by amount of value in Illusion fleet.
    float capRate = 0.25 * std::min(1.0f, std::sqrt(300000.0f / 200000));
    EXPECT_TRUE(capRate == 0.25);
    NumberOfShips mayCap =  500 * capRate;
    NumberOfShips canCap0 = (700 * stats["Illusion"]->guns + 30 * stats["Ziz"]->guns     ) / stats["Asteroid"]->armor;
    NumberOfShips canCap1 = (400 * stats["Illusion"]->guns + 80 * stats["Merchant"]->guns) / stats["Asteroid"]->armor;
    EXPECT_TRUE(canCap0 + canCap1 > mayCap);
    NumberOfShips value0 = 700 * stats["Illusion"]->cost + 30 * stats["Ziz"]->cost;
    NumberOfShips value1 = 400 * stats["Illusion"]->cost + 80 * stats["Merchant"]->cost;

    NumberOfShips cap0 = mayCap * value0 / (value0 + value1);
    NumberOfShips cap1 = mayCap * value1 / (value0 + value1);

    expect.fleet(Side::Attack,  0)->addStolen(*stats["Asteroid"], cap0);
    expect.fleet(Side::Attack,  1)->addStolen(*stats["Asteroid"], cap1);
    float bravery0 = std::max(0.2f, std::min(2.2f,  float(expect.fleet(Side::Defense, 0)->score) / expect.fleet(Side::Attack, 0)->score) - 0.2f) *
        std::max(0.2f, std::min(1.8f,  float(expect.fleet(Side::Defense, 0)->value) / expect.fleet(Side::Attack, 0)->value) - 0.1f) /
        std::max(1.0f,          0.6f + float(expect.fleet(Side::Attack,  0)->score) / expect.fleet(Side::Attack, 0)->value / 10);
    float bravery1 = std::max(0.2f, std::min(2.2f,  float(expect.fleet(Side::Defense, 0)->score) / expect.fleet(Side::Attack, 1)->score) - 0.2f) *
        std::max(0.2f, std::min(1.8f,  float(expect.fleet(Side::Defense, 0)->value) / expect.fleet(Side::Attack, 1)->value) - 0.1f) /
        std::max(1.0f,          0.6f + float(expect.fleet(Side::Attack,  1)->score) / expect.fleet(Side::Attack, 1)->value / 10);
    expect.fleet(Side::Attack,  0)->addXP(cap0 * 10 * bravery0);
    expect.fleet(Side::Attack,  1)->addXP(cap1 * 10 * bravery1);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Asteroid")->cap(cap0 + cap1), cap0 + cap1);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void xp() {
    // Enough Illusion to maxcap on target.
    Battle actual(1, 1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Illusion"], 1500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  500));
    actual.fleet(Side::Attack,  0)->value = 400000;
    actual.fleet(Side::Attack,  0)->score = 600000;
    actual.fleet(Side::Attack,  0)->militaryCenters = 199;
    actual.fleet(Side::Defense, 0)->value = 500000;
    actual.fleet(Side::Defense, 0)->score = 800000;

    Battle expect(actual);
    NumberOfShips canCap = 1500 * stats["Illusion"]->guns / stats["Asteroid"]->armor;
    NumberOfShips mayCap = 500 * 0.25 * std::min(1.0f, std::sqrt(500000.0f / 400000));
    EXPECT_TRUE(canCap >= mayCap);
    expect.fleet(Side::Attack,  0)->addStolen(*stats["Asteroid"], mayCap);
    float bravery = std::max(0.2f, std::min(2.2f, float(expect.fleet(Side::Defense, 0)->score) / expect.fleet(Side::Attack, 0)->score) - 0.2f) *
        std::max(0.2f, std::min(1.8f, float(expect.fleet(Side::Defense, 0)->value) / expect.fleet(Side::Attack,  0)->value) - 0.1f) /
        std::max(1.0f, 0.6f + float(expect.fleet(Side::Attack,  0)->score) / expect.fleet(Side::Attack,  0)->value / 10);
    expect.fleet(Side::Attack,  0)->addXP(mayCap * 10 * bravery * (1 + float(199) / 200));
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Asteroid")->cap(canCap), mayCap);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}


// STRUCTURE KILL //////////////////////////////////////////////////////////////

void structureKill() {
    Battle actual(1,1);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Behemoth"],   404));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Structure"], 1000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],     0));

    Battle expect(actual);
    NumberOfShips mayDestroy = 0.1 * 1000;
    NumberOfShips canDestroy = 404 * stats["Behemoth"]->guns / stats["Structure"]->armor;
    EXPECT_TRUE(canDestroy > mayDestroy);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Structure")->structureKill(canDestroy), mayDestroy);

    NumberOfShips abandon = mayDestroy * stats["Structure"]->armor / stats["Behemoth"]->guns;
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Behemoth")->kill(abandon), abandon);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}

void distributeStructureKill() {
    Battle actual(2,2);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Behemoth"],  102));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  1, *stats["Behemoth"],  306));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Structure"], 500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 1, *stats["Structure"], 500));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],  0));

    Battle expect(actual);
    NumberOfShips mayDestroy = 0.1 * 500;
    NumberOfShips canDestroy = 408 / 2 * stats["Behemoth"]->guns / stats["Structure"]->armor;
    EXPECT_TRUE(canDestroy > mayDestroy);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 0)->battleFormation("Structure")->structureKill(canDestroy), mayDestroy);
    EXPECT_EQUAL(expect.fleet(Side::Defense, 1)->battleFormation("Structure")->structureKill(canDestroy), mayDestroy);

    NumberOfShips abandon0 = 2 * mayDestroy * stats["Structure"]->armor / stats["Behemoth"]->guns / 4;
    NumberOfShips abandon1 = 6 * mayDestroy * stats["Structure"]->armor / stats["Behemoth"]->guns / 4;
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Behemoth")->kill(abandon0), abandon0);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 1)->battleFormation("Behemoth")->kill(abandon1), abandon1);

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}


// SALVAGE /////////////////////////////////////////////////////////////////////

void raceSalvage() {
    // Phantom and Corsair do not fire at Pegasus fleets. Pegasus fire at
    // Phantom and Corsair.
    Battle actual(1, 6);
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Phantom"],  20000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Attack,  0, *stats["Corsair"],  10000));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],     0));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 0, *stats["Pegasus"],    150));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 1, *stats["Pegasus"],    150));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 2, *stats["Pegasus"],    150));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 3, *stats["Pegasus"],    150));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 4, *stats["Pegasus"],    150));
    EXPECT_TRUE(actual.addBattleFormation(Side::Defense, 5, *stats["Pegasus"],    150));
    actual.fleet(Side::Defense, 0)->race = Race::Ter;
    actual.fleet(Side::Defense, 1)->race = Race::Cat;
    actual.fleet(Side::Defense, 2)->race = Race::Xan;
    actual.fleet(Side::Defense, 3)->race = Race::Zik;
    actual.fleet(Side::Defense, 4)->race = Race::Etd;
    actual.fleet(Side::Defense, 5)->race = Race::None;

    Battle expect(actual);
    NumberOfShips deadPhantom =
        900 * 20000 / (20000+10000) * stats["Pegasus"]->guns / stats["Phantom"]->armor;
    NumberOfShips deadCorsair =
        900 * 10000 / (20000+10000) * stats["Pegasus"]->guns / stats["Corsair"]->armor;
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Phantom")->kill(deadPhantom), deadPhantom);
    EXPECT_EQUAL(expect.fleet(Side::Attack, 0)->battleFormation("Corsair")->kill(deadCorsair), deadCorsair);
    Cost salvage = (deadCorsair * stats["Corsair"]->cost +
                    deadPhantom * stats["Phantom"]->cost) * 0.2 / 6;
    Cost salvageCap = expect.fleet(Side::Defense, 0)->fleetCost() / 5;
    EXPECT_TRUE(salvage > salvageCap);
    expect.fleet(Side::Defense, 0)->grantSalvage(std::ceil(float(salvageCap) / Race(Race::Ter).salvageFactor));
    expect.fleet(Side::Defense, 1)->grantSalvage(std::ceil(float(salvageCap) / Race(Race::Cat).salvageFactor));
    expect.fleet(Side::Defense, 2)->grantSalvage(std::ceil(float(salvageCap) / Race(Race::Xan).salvageFactor));
    expect.fleet(Side::Defense, 3)->grantSalvage(std::ceil(float(salvageCap) / Race(Race::Zik).salvageFactor));
    expect.fleet(Side::Defense, 4)->grantSalvage(std::ceil(float(salvageCap) / Race(Race::Etd).salvageFactor));
    expect.fleet(Side::Defense, 5)->grantSalvage(std::ceil(float(salvageCap) / Race(Race::None).salvageFactor));

    bcalc(actual, false);
    EXPECT_EQUAL(actual, expect);
}


// ANALYZE BATTLE //////////////////////////////////////////////////////////////

namespace {
void setupBattleForLandable(Battle& battle, float ratio, float repayTicks) {
    NumberOfShips deadZiz = repayTicks * 250 * 200 * 1.64 / stats["Ziz"]->cost;
    NumberOfShips startHarpy = std::ceil(float(deadZiz * stats["Ziz"]->armor) / stats["Harpy"]->guns);
    NumberOfShips deadBuccaneer = ratio * deadZiz * stats["Ziz"]->cost / stats["Buccaneer"]->cost;
    NumberOfShips startWyvern = deadBuccaneer * stats["Buccaneer"]->armor / stats["Wyvern"]->guns;
    battle.addBattleFormation(Side::Attack,  0, *stats["Ziz"],          10000); // Killed by Harpy
    battle.addBattleFormation(Side::Attack,  0, *stats["Wyvern"], startWyvern); // Kills Buccaneer
    battle.addBattleFormation(Side::Defense, 0, *stats["Ziz"],              1);
    battle.addBattleFormation(Side::Defense, 0, *stats["Asteroid"],       800);
    battle.addBattleFormation(Side::Defense, 0, *stats["Harpy"],   startHarpy); // Kills Ziz
    battle.addBattleFormation(Side::Defense, 0, *stats["Buccaneer"],   100000); // Killed by Wyvern
    bcalc(battle, false);
    battle.fleet(Side::Defense, 0)->salvage = 0; // Disregard salvage.
}
}

void landableForLossless() {
    // Empty.
    Battle battle(1, 1);
    battle.addBattleFormation(Side::Attack,  0, *stats["Ziz"],     1000);
    battle.addBattleFormation(Side::Defense, 0, *stats["Ziz"],        1);
    battle.addBattleFormation(Side::Defense, 0, *stats["Asteroid"], 800);
    bcalc(battle, false);
    EXPECT_TRUE(battle.landableFor(Side::Attack));
}

void landableFor14() {
    Battle battle(1, 1);
    setupBattleForLandable(battle, 0.6f, 14.55f);
    EXPECT_TRUE(battle.landableFor(Side::Attack));
}

void landableFor16() {
    Battle battle(1, 1);
    setupBattleForLandable(battle, 1.4f, 16.60f);
    EXPECT_TRUE(battle.landableFor(Side::Attack));
}

void landableFor31() {
    Battle battle(1, 1);
    setupBattleForLandable(battle, 2.4f, 31.20f);
    EXPECT_TRUE(battle.landableFor(Side::Attack));
}

void landableFor80() {
    Battle battle(1, 1);
    setupBattleForLandable(battle, 3.4f, 80.68f);
    EXPECT_TRUE(battle.landableFor(Side::Attack));
}

void notLandableFor12() {
    Battle battle(1, 1);
    setupBattleForLandable(battle, 0.6f, 14.6f);
    EXPECT_TRUE(!battle.landableFor(Side::Attack));
}

void notLandableFor24() {
    Battle battle(1, 1);
    setupBattleForLandable(battle, 1.4f, 16.62f);
    EXPECT_TRUE(!battle.landableFor(Side::Attack));
}

void notLandableFor48() {
    Battle battle(1, 1);
    setupBattleForLandable(battle, 2.4f, 31.45f);
    EXPECT_TRUE(!battle.landableFor(Side::Attack));
}

void notLandableFor96() {
    Battle battle(1, 1);
    setupBattleForLandable(battle, 3.4f, 80.88f);
    EXPECT_TRUE(!battle.landableFor(Side::Attack));
}

void analyzeBattle0v0() {
    // Just pods roiding a defenseless planet.
    Fleet attacker(0, Side::Attack);
    attacker.addBattleFormation(*stats["Ziz"],     2000);
    Fleet defender(0, Side::Defense);
    defender.addBattleFormation(*stats["Ziz"],        1);
    defender.addBattleFormation(*stats["Asteroid"], 800);

    AnalysisOutcome actual = analyzeBattle(stats,
                                           attacker, defender,
                                           valueRatio,
                                           withoutRepeatSameTargetIndex, withoutFlak);
    EXPECT_TRUE(actual.landable);
    EXPECT_TRUE(!actual.hasCounter);
}

void analyzeBattle1v0() {
    // Pods plus 1 escort roiding a defenseless planet.
    Fleet attacker(0, Side::Attack);
    attacker.addBattleFormation(*stats["Ziz"],     2000);
    attacker.addBattleFormation(*stats["Pegasus"], 2000);
    Fleet defender(0, Side::Defense);
    defender.addBattleFormation(*stats["Ziz"],        1);
    defender.addBattleFormation(*stats["Asteroid"], 800);

    AnalysisOutcome actual = analyzeBattle(stats,
                                           attacker, defender,
                                           valueRatio,
                                           withoutRepeatSameTargetIndex, withoutFlak);
    EXPECT_TRUE(actual.landable);
    EXPECT_TRUE(!actual.hasCounter);
}

void analyzeBattle1v1Wrong() {
    // Pods plus 1 escort roiding a planet with the wrong defense ship.
    Fleet attacker(0, Side::Attack);
    attacker.addBattleFormation(*stats["Ziz"],       2000);
    attacker.addBattleFormation(*stats["Pegasus"],   2000);
    Fleet defender(0, Side::Defense);
    defender.addBattleFormation(*stats["Ziz"],          1);
    defender.addBattleFormation(*stats["Asteroid"],   800);
    defender.addBattleFormation(*stats["Phantom"], 100000);

    AnalysisOutcome actual = analyzeBattle(stats,
                                           attacker, defender,
                                           valueRatio,
                                           withoutRepeatSameTargetIndex, withoutFlak);
    EXPECT_TRUE(actual.landable);
    EXPECT_TRUE(!actual.hasCounter);
}

void analyzeBattle1v1LosslessLand() {
    // Pods plus 1 escort roiding a planet with too few untargetted defense ships.
    Fleet attacker(0, Side::Attack);
    attacker.addBattleFormation(*stats["Ziz"],     2000);
    attacker.addBattleFormation(*stats["Pegasus"], 2000);
    Fleet defender(0, Side::Defense);
    defender.addBattleFormation(*stats["Ziz"],        1);
    defender.addBattleFormation(*stats["Asteroid"], 800);
    defender.addBattleFormation(*stats["Dealer"],   223);

    AnalysisOutcome actual = analyzeBattle(stats,
                                           attacker, defender,
                                           valueRatio,
                                           withoutRepeatSameTargetIndex, withoutFlak);
    EXPECT_TRUE(actual.landable);
    EXPECT_TRUE(actual.hasCounter);
}

void analyzeBattle1v1LosslessRecall() {
    // Pods plus 1 escort roiding a planet with plenty of untargetted defense ships.
    Fleet attacker(0, Side::Attack);
    attacker.addBattleFormation(*stats["Ziz"],     2000);
    attacker.addBattleFormation(*stats["Pegasus"], 2000);
    Fleet defender(0, Side::Defense);
    defender.addBattleFormation(*stats["Ziz"],        1);
    defender.addBattleFormation(*stats["Asteroid"], 800);
    defender.addBattleFormation(*stats["Dealer"],   224);

    AnalysisOutcome actual = analyzeBattle(stats,
                                           attacker, defender,
                                           valueRatio,
                                           withoutRepeatSameTargetIndex, withoutFlak);
    EXPECT_TRUE(!actual.landable);
    EXPECT_TRUE(actual.hasCounter);
}

void analyzeBattle1v1Land() {
    // Pods plus 1 escort roiding a planet with way too few targetted defense
    // ships.
    Fleet attacker(0, Side::Attack);
    attacker.addBattleFormation(*stats["Ziz"],     2000);
    attacker.addBattleFormation(*stats["Pegasus"], 2000);
    Fleet defender(0, Side::Defense);
    defender.addBattleFormation(*stats["Ziz"],        1);
    defender.addBattleFormation(*stats["Asteroid"], 800);
    defender.addBattleFormation(*stats["Harpy"],   2984);

    AnalysisOutcome actual = analyzeBattle(stats,
                                           attacker, defender,
                                           valueRatio,
                                           withoutRepeatSameTargetIndex, withoutFlak);
    EXPECT_TRUE(actual.landable);
    EXPECT_TRUE(actual.hasCounter);
}

void analyzeBattle1v1Recall() {
    // Pods plus 1 escort roiding a planet with plenty of targetted defense ships.
    Fleet attacker(0, Side::Attack);
    attacker.addBattleFormation(*stats["Ziz"],     2000);
    attacker.addBattleFormation(*stats["Pegasus"], 2000);
    Fleet defender(0, Side::Defense);
    defender.addBattleFormation(*stats["Ziz"],        1);
    defender.addBattleFormation(*stats["Asteroid"], 800);
    defender.addBattleFormation(*stats["Harpy"],   2985);

    AnalysisOutcome actual = analyzeBattle(stats,
                                           attacker, defender,
                                           valueRatio,
                                           withoutRepeatSameTargetIndex, withoutFlak);
    EXPECT_TRUE(!actual.landable);
    EXPECT_TRUE(actual.hasCounter);
}

void analyzeBattle1vChoose() {
    // Pods plus 1 escort roiding a planet with too few targetted defense ships
    // of one type, and plenty of untargetted defense ships of another.
    Fleet attacker(0, Side::Attack);
    attacker.addBattleFormation(*stats["Ziz"],     2000);
    attacker.addBattleFormation(*stats["Pegasus"], 2000);
    Fleet defender(0, Side::Defense);
    defender.addBattleFormation(*stats["Ziz"],        1);
    defender.addBattleFormation(*stats["Asteroid"], 800);
    defender.addBattleFormation(*stats["Harpy"],    100);
    defender.addBattleFormation(*stats["Dealer"],  1000);

    AnalysisOutcome actual = analyzeBattle(stats,
                                           attacker, defender,
                                           valueRatio,
                                           withoutRepeatSameTargetIndex, withoutFlak);
    EXPECT_TRUE(!actual.landable);
    EXPECT_TRUE(actual.hasCounter);
}

void analyzeBattle1v2Land() {
    // Pods plus 1 escort roiding a planet with too few targetted defense ships
    // of one type, and too few untargetted defense ships of another. Together
    // it's also not enough.
    Fleet attacker(0, Side::Attack);
    attacker.addBattleFormation(*stats["Ziz"],     2000);
    attacker.addBattleFormation(*stats["Pegasus"], 2000);
    Fleet defender(0, Side::Defense);
    defender.addBattleFormation(*stats["Ziz"],        1);
    defender.addBattleFormation(*stats["Asteroid"], 800);
    defender.addBattleFormation(*stats["Harpy"],   1200);
    defender.addBattleFormation(*stats["Dealer"],   120);

    AnalysisOutcome actual = analyzeBattle(stats,
                                           attacker, defender,
                                           valueRatio,
                                           withoutRepeatSameTargetIndex, withoutFlak);
    EXPECT_TRUE(actual.landable);
    EXPECT_TRUE(actual.hasCounter);
}

void analyzeBattle1v2Recall() {
    // Pods plus 1 escort roiding a planet with too few targetted defense ships
    // of one type, and too few untargetted defense ships of another. Together
    // it's enough, though.
    Fleet attacker(0, Side::Attack);
    attacker.addBattleFormation(*stats["Ziz"],     2000);
    attacker.addBattleFormation(*stats["Pegasus"], 2000);
    Fleet defender(0, Side::Defense);
    defender.addBattleFormation(*stats["Ziz"],        1);
    defender.addBattleFormation(*stats["Asteroid"], 800);
    defender.addBattleFormation(*stats["Harpy"],   2000);
    defender.addBattleFormation(*stats["Dealer"],   150);

    AnalysisOutcome actual = analyzeBattle(stats,
                                           attacker, defender,
                                           valueRatio,
                                           withoutRepeatSameTargetIndex, withoutFlak);
    EXPECT_TRUE(!actual.landable);
    EXPECT_TRUE(actual.hasCounter);
}

void analyzeBattleFlak() {
    Fleet attacker(0, Side::Attack);
    attacker.addBattleFormation(*stats["Ziz"],     2000);
    attacker.addBattleFormation(*stats["Pegasus"], 2000);
    Fleet defender(0, Side::Defense);
    defender.addBattleFormation(*stats["Ziz"],        1);
    defender.addBattleFormation(*stats["Asteroid"], 800);
    defender.addBattleFormation(*stats["Phoenix"], 7200);
    defender.addBattleFormation(*stats["Beetle"],  7200);

    AnalysisOutcome actual = analyzeBattle(stats,
                                           attacker, defender,
                                           valueRatio,
                                           withoutRepeatSameTargetIndex, withoutFlak);
    EXPECT_TRUE(actual.landable);
    EXPECT_TRUE(actual.hasCounter);

    AnalysisOutcome actualFlak = analyzeBattle(stats,
                                               attacker, defender,
                                               valueRatio,
                                               withoutRepeatSameTargetIndex, withFlak);
    EXPECT_TRUE(!actualFlak.landable);
    EXPECT_TRUE(actualFlak.hasCounter);
}


// MAYBE ///////////////////////////////////////////////////////////////////////

void maybeNothing() {
    Maybe<GunType> nothingGun(nullptr);
    EXPECT_TRUE(!nothingGun);
    // Dereferencing an invalid Maybe results in a segmentation fault. So don't
    // do that.
}

void maybeSomething() {
    GunType type(GunType::Kill);
    Maybe<GunType> justGun(&type);
    EXPECT_TRUE(bool(justGun));
    EXPECT_EQUAL(*justGun, type);
}

void maybeNothingExec() {
    Maybe<ShipType> nothingShip(nullptr);
    // Calling exec() on an invalid Maybe results in another invalid Maybe. The
    // function passed in is not executed.
    Maybe<Class> nothingClass = nothingShip.exec<Class>(&ShipType::target, std::size_t(0));
    EXPECT_TRUE(!nothingClass);

    // Calls to exec() can therefore safely be chained.
    Maybe<std::string> nothingString = nothingShip
        .exec<Class>(&ShipType::target, std::size_t(2))
        .exec<std::string>(&Class::toString)
        .exec<std::string>(&std::string::substr, std::size_t(0), std::size_t(3));
    EXPECT_TRUE(!nothingString);
}

void maybeNothingConstExec() {
    const Maybe<const ShipType> nothingShip(nullptr);
    // Calling exec on an invalid const Maybe results in another invalid Maybe.
    // The function pointed to is not executed.
    Maybe<Class> nothingClass = nothingShip.exec<Class>(&ShipType::target, std::size_t(0));
    EXPECT_TRUE(!nothingClass);

    // And these can also be chained.
    Maybe<std::string> nothingString = nothingShip
        .exec<Class>(&ShipType::target, std::size_t(2))
        .exec<std::string>(&Class::toString)
        .exec<std::string>(&std::string::substr, std::size_t(0), std::size_t(3));
    EXPECT_TRUE(!nothingString);
}

void maybeSomethingExec() {
    ShipType s;
    s.setTarget(0, Class::None);
    s.setTarget(1, Class::Destroyer);
    s.setTarget(2, Class::Cruiser);
    Maybe<ShipType> justShip(&s);
    // Calling exec on a valid Maybe results in another valid Maybe, that being
    // the result of the function pointed to.
    Maybe<Class> justClass = justShip.exec<Class>(&ShipType::target, std::size_t(1));
    EXPECT_EQUAL(*justClass, Class(Class::Destroyer));

    Maybe<Class> noneClass = justShip.exec<Class>(&ShipType::target, std::size_t(0));
    EXPECT_EQUAL(*noneClass, Class(Class::None));

    Maybe<std::string> justString = justShip
        .exec<Class>(&ShipType::target, std::size_t(2))
        .exec<std::string>(&Class::toString)
        .exec<std::string>(&std::string::substr, std::size_t(0), std::size_t(2));
    EXPECT_EQUAL(*justString, std::string("Cr"));
}

void maybeSomethingConstExec() {
    ShipType s;
    s.setTarget(0, Class::Fighter);
    const Maybe<const ShipType> justShip(&s);
    // We have a const exec too. The constness of the type inside is ignored by
    // Maybe, but taken into account in the call to the passed function.
    Maybe<Class> justClass = justShip.exec<Class>(&ShipType::target, std::size_t(0));
    EXPECT_EQUAL(*justClass, Class(Class::Fighter));
}

void maybeNothingMember() {
    Maybe<ShipType> nothingShip(nullptr);

    // Fields can be accessed in the same way.
    Maybe<Guns> nothingGuns = nothingShip.get<Guns>(&ShipType::guns);
    EXPECT_TRUE(!nothingGuns);
    Maybe<Armor> nothingArmor = nothingShip.get<Armor>(&ShipType::armor);
    EXPECT_TRUE(!nothingArmor);
}

void maybeNothingConstMember() {
    const Maybe<const ShipType> nothingShip(nullptr);
    Maybe<ERes> nothingGuns = nothingShip.get<ERes>(&ShipType::eres);
    EXPECT_TRUE(!nothingGuns);
}

void maybeSomethingMember() {
    ShipType s;
    s.guns = 99;
    s.armor = 98;
    Maybe<ShipType> justShip(&s);
    Maybe<Guns> justGuns = justShip.get<Guns>(&ShipType::guns);
    EXPECT_EQUAL(*justGuns, Guns(99));
    Maybe<Armor> justArmor = justShip.get<Armor>(&ShipType::armor);
    EXPECT_EQUAL(*justArmor, Armor(98));
}

void maybeSomethingConstMember() {
    ShipType s;
    s.eres = 97;

    const Maybe<const ShipType> constShip(&s);
    Maybe<ERes> constGuns = constShip.get<ERes>(&ShipType::eres);
    EXPECT_EQUAL(*constGuns, ERes(97));
}


// MAIN ////////////////////////////////////////////////////////////////////////

void readTestStats(Stats& stats, Verbosity verbosity = Verbosity::Warning) {
    Logger log(verbosity);

    EXPECT_TRUE(stats.create("Asteroid",
                             Class::Asteroid,
                             std::array<Class, MAX_TARGETS>{},
                             GunType::None,
                             Init(0),
                             Guns(0),
                             Armor(50),
                             ERes(0),
                             Cost(20000),
                             Race::None));

    EXPECT_TRUE(stats.create("Structure",
                             Class::Structure,
                             std::array<Class, MAX_TARGETS>{},
                             GunType::None,
                             Init(0),
                             Guns(0),
                             Armor(500),
                             ERes(0),
                             Cost(20000),
                             Race::None));


    EXPECT_TRUE(stats.create("Harpy",
                             Class::Fighter,
                             std::array<Class, MAX_TARGETS>{ Class::Destroyer },
                             GunType::Kill,
                             Init(3),
                             Guns(4),
                             Armor(5),
                             ERes(55),
                             Cost(500),
                             Race::Ter));

    EXPECT_TRUE(stats.create("Phoenix",
                             Class::Corvette,
                             std::array<Class, MAX_TARGETS>{ Class::Destroyer, Class::Frigate },
                             GunType::Kill,
                             Init(6),
                             Guns(8),
                             Armor(10),
                             ERes(75),
                             Cost(1000),
                             Race::Ter));

    EXPECT_TRUE(stats.create("Pegasus",
                             Class::Destroyer,
                             std::array<Class, MAX_TARGETS>{ Class::Fighter, Class::Corvette },
                             GunType::Kill,
                             Init(5),
                             Guns(60),
                             Armor(50),
                             ERes(95),
                             Cost(5000),
                             Race::Ter));

    EXPECT_TRUE(stats.create("Wyvern",
                             Class::Battleship,
                             std::array<Class, MAX_TARGETS>{ Class::Frigate },
                             GunType::Kill,
                             Init(7),
                             Guns(400),
                             Armor(560),
                             ERes(99),
                             Cost(40000),
                             Race::Ter));

    EXPECT_TRUE(stats.create("Behemoth",
                             Class::Cruiser,
                             std::array<Class, MAX_TARGETS>{ Class::Structure },
                             GunType::StructureKill,
                             Init(50),
                             Guns(125),
                             Armor(150),
                             ERes(96),
                             Cost(20000),
                             Race::Ter));

    EXPECT_TRUE(stats.create("Ziz",
                             Class::Destroyer,
                             std::array<Class, MAX_TARGETS>{ Class::Asteroid },
                             GunType::Pod,
                             Init(40),
                             Guns(50),
                             Armor(30),
                             ERes(92),
                             Cost(3000),
                             Race::Ter));

    EXPECT_TRUE(stats.create("Vindicator",
                             Class::Fighter,
                             std::array<Class, MAX_TARGETS>{ Class::Frigate },
                             GunType::Kill,
                             Init(11),
                             Guns(10),
                             Armor(10),
                             ERes(37),
                             Cost(295),
                             Race::Cat));

    EXPECT_TRUE(stats.create("Spider",
                             Class::Fighter,
                             std::array<Class, MAX_TARGETS>{ Class::Fighter },
                             GunType::EMP,
                             Init(1),
                             Guns(1),
                             Armor(6),
                             ERes(40),
                             Cost(300),
                             Race::Cat));

    EXPECT_TRUE(stats.create("Viper",
                             Class::Fighter,
                             std::array<Class, MAX_TARGETS>{ Class::Corvette, Class::Frigate },
                             GunType::EMP,
                             Init(3),
                             Guns(2),
                             Armor(6),
                             ERes(40),
                             Cost(300),
                             Race::Cat));

    EXPECT_TRUE(stats.create("Beetle",
                             Class::Corvette,
                             std::array<Class, MAX_TARGETS>{ Class::Fighter },
                             GunType::EMP,
                             Init(2),
                             Guns(4),
                             Armor(6),
                             ERes(65),
                             Cost(600),
                             Race::Cat));

    EXPECT_TRUE(stats.create("Phantom",
                             Class::Fighter,
                             std::array<Class, MAX_TARGETS>{ Class::Frigate },
                             GunType::Kill,
                             Init(3),
                             Guns(4),
                             Armor(3),
                             ERes(10),
                             Cost(300),
                             Race::Xan));

    EXPECT_TRUE(stats.create("Spectre",
                             Class::Destroyer,
                             std::array<Class, MAX_TARGETS>{ Class::Fighter },
                             GunType::Kill,
                             Init(5),
                             Guns(30),
                             Armor(25),
                             ERes(90),
                             Cost(2500),
                             Race::Xan));

    EXPECT_TRUE(stats.create("Illusion",
                             Class::Fighter,
                             std::array<Class, MAX_TARGETS>{ Class::Asteroid },
                             GunType::Pod,
                             Init(40),
                             Guns(5),
                             Armor(3),
                             ERes(20),
                             Cost(300),
                             Race::Xan));

    EXPECT_TRUE(stats.create("Corsair",
                             Class::Fighter,
                             std::array<Class, MAX_TARGETS>{ Class::Corvette },
                             GunType::Kill,
                             Init(5),
                             Guns(5),
                             Armor(5),
                             ERes(40),
                             Cost(400),
                             Race::Zik));

    EXPECT_TRUE(stats.create("Thief",
                             Class::Fighter,
                             std::array<Class, MAX_TARGETS>{ Class::Corvette },
                             GunType::Steal,
                             Init(23),
                             Guns(2),
                             Armor(5),
                             ERes(30),
                             Cost(300),
                             Race::Zik));

    EXPECT_TRUE(stats.create("Cutlass",
                             Class::Corvette,
                             std::array<Class, MAX_TARGETS>{ Class::Fighter, Class::Corvette },
                             GunType::Steal,
                             Init(21),
                             Guns(6),
                             Armor(7),
                             ERes(70),
                             Cost(650),
                             Race::Zik));

    EXPECT_TRUE(stats.create("Cutter",
                             Class::Corvette,
                             std::array<Class, MAX_TARGETS>{ Class::Fighter },
                             GunType::Steal,
                             Init(23),
                             Guns(4),
                             Armor(9),
                             ERes(70),
                             Cost(650),
                             Race::Zik));

    EXPECT_TRUE(stats.create("Buccaneer",
                             Class::Frigate,
                             std::array<Class, MAX_TARGETS>{ Class::Fighter },
                             GunType::Kill,
                             Init(22),
                             Guns(17),
                             Armor(15),
                             ERes(87),
                             Cost(1500),
                             Race::Zik));

    EXPECT_TRUE(stats.create("Clipper",
                             Class::Destroyer,
                             std::array<Class, MAX_TARGETS>{ Class::Fighter },
                             GunType::Kill,
                             Init(44),
                             Guns(40),
                             Armor(39),
                             ERes(94),
                             Cost(3000),
                             Race::Zik));

    EXPECT_TRUE(stats.create("Ranger",
                             Class::Frigate,
                             std::array<Class, MAX_TARGETS>{ Class::Battleship },
                             GunType::Kill,
                             Init(7),
                             Guns(28),
                             Armor(26),
                             ERes(80),
                             Cost(2000),
                             Race::Etd));

    EXPECT_TRUE(stats.create("Merchant",
                             Class::Destroyer,
                             std::array<Class, MAX_TARGETS>{ Class::Asteroid },
                             GunType::Pod,
                             Init(41),
                             Guns(40),
                             Armor(24),
                             ERes(90),
                             Cost(2400),
                             Race::Etd));

    EXPECT_TRUE(stats.create("Avenger",
                             Class::Frigate,
                             std::array<Class, MAX_TARGETS>{ Class::Fighter },
                             GunType::Kill,
                             Init(11),
                             Guns(49),
                             Armor(47),
                             ERes(45),
                             Cost(1122),
                             Race::Etd));

    EXPECT_TRUE(stats.create("Dealer",
                             Class::Frigate,
                             std::array<Class, MAX_TARGETS>{ Class::Destroyer },
                             GunType::Kill,
                             Init(10),
                             Guns(40),
                             Armor(40),
                             ERes(40),
                             Cost(1000),
                             Race::Etd));

    log.info("Read %lu ship types\n\n",
             stats.size());
}


int main(int /*argc*/, char** /*argv*/) {
    RESET_ERRORS();

    readTestStats(stats);

    emptyBattle();
    emptyFleet();
    emptyFleets();
    nonEmptyAttack();
    noInteraction();

    zeroLossKill();
    distributeKill();
    dontTargetDeadShips();
    exchangeKill();
    sameTimeKill();
    sameTimeFullKill();
    overshootKill();
    t2Kill();
    t2AfterOvershootKill();

    zeroLossEmp();
    distributeEmp();
    exchangeEmp();
    sameTimeEmp();
    overshootEmp();
    t2Emp();
    t2AfterOvershootEmp();

    zeroLossSteal();
    stolenShipsUntargetableSteal();
    distributeSteal();
    exchangeSteal();
    sameTimeSteal();
    overshootSteal();
    t2Steal();
    t2AfterOvershootSteal();

    maxCap();
    notEnoughPodsCap();
    deadPodsCantCap();
    stillAlivePodsCap();
    emptyFleetsCantCap();
    multipleTypesCap();
    variableLimitCap();
    distributeCap();
    xp();

    structureKill();
    distributeStructureKill();

    raceSalvage();

    overshootRepeatKill();
    t2AfterRepeatOvershootKill();
    overshootRepeatEmp();
    t2AfterRepeatOvershootEmp();
    overshootRepeatSteal();
    t2AfterRepeatOvershootSteal();

    landableForLossless();
    landableFor14();
    landableFor16();
    landableFor31();
    landableFor80();
    notLandableFor12();
    notLandableFor24();
    notLandableFor48();
    notLandableFor96();

    analyzeBattle0v0();
    analyzeBattle1v0();
    analyzeBattle1v1Wrong();
    analyzeBattle1v1LosslessLand();
    analyzeBattle1v1LosslessRecall();
    analyzeBattle1v1Land();
    analyzeBattle1v1Recall();
    analyzeBattle1vChoose();
    analyzeBattle1v2Land();
    analyzeBattle1v2Recall();

    analyzeBattleFlak();

    maybeNothing();
    maybeSomething();
    maybeNothingExec();
    maybeNothingConstExec();
    maybeSomethingExec();
    maybeSomethingConstExec();
    maybeNothingMember();
    maybeNothingConstMember();
    maybeSomethingMember();
    maybeSomethingConstMember();

    int errors = GET_ERROR_COUNT();
    if (errors) {
        printf("%d test failures!\n", errors);
    }
    else {
        printf("All tests passed!\n");
    }
    return !!errors;
}
